# Author: Carsten Sachse 08-Jun-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
from collections import namedtuple
import os
import shutil
from spring.csinfrastr.csdatabase import SpringDataBase, refine_base, base, RefinementCycleTable, \
    RefinementCycleHelixTable, RefinementCycleSegmentTable, SegmentTable, HelixTable, CtfMicrographTable
from spring.segment2d.segmentselect import SegmentSelect
from spring.segment3d.refine.sr3d_align import SegmentRefine3dAlign
from spring.segment3d.segclassreconstruct import SegClassReconstruct

from sqlalchemy.sql.expression import or_, desc, and_

import numpy as np


class SegmentRefine3dSelectionFilter(SegmentRefine3dAlign):
    def get_excluded_refinement_count(self, session, included_segments_classes):
        segment_count = session.query(SegmentTable).order_by(SegmentTable.stack_id).count()
        excluded_count = segment_count - len(set(included_segments_classes))
        
        return excluded_count
    
    
    def filter_refined_segments_by_property(self, session, ref_session, refined_segment_table_property, last_cycle,
    property_selection, property_in_or_exclude, property_range):
        included_segments_property = []
        if property_selection:
            if property_in_or_exclude == 'include':
                included_segments = ref_session.query(RefinementCycleSegmentTable).\
                filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
                filter(and_(refined_segment_table_property >= property_range[0], 
                            refined_segment_table_property <=property_range[1])).all()
                
                included_segments_property = [each_segment.stack_id for each_segment in included_segments]

            elif property_in_or_exclude == 'exclude':
                included_segments = ref_session.query(RefinementCycleSegmentTable).\
                filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
                filter(or_(refined_segment_table_property < property_range[0], 
                           refined_segment_table_property > property_range[1])).all()
                           
                included_segments_property = [each_segment.stack_id for each_segment in included_segments]

        elif not property_selection:
            included_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).all()
            included_segments_property = [each_segment.stack_id for each_segment in included_segments]
            
        excluded_segment_count = self.get_excluded_refinement_count(session, included_segments_property)
        
        return included_segments_property, excluded_segment_count
    
    
    def filter_segments_by_ccc_against_projections(self, session, ref_session, last_cycle, ccc_proj_selection,
    ccc_proj_in_or_exclude, ccc_proj_range):
        included_segments_ccc_proj, excluded_proj_cc_count = self.filter_refined_segments_by_property(session,
        ref_session, RefinementCycleSegmentTable.peak, last_cycle, ccc_proj_selection, ccc_proj_in_or_exclude,
        ccc_proj_range)
                    
        return included_segments_ccc_proj, excluded_proj_cc_count
    
    
    def filter_segments_by_out_of_plane_tilt(self, session, ref_session, last_cycle, out_of_plane_selection,
    out_of_plane_in_or_exclude, out_of_plane_in_or_ex_range):
        
        included_segments_oop_tilt, excluded_segments_oop_tilt_count = \
        self.filter_refined_segments_by_property(session, ref_session, RefinementCycleSegmentTable.theta - 90,
        last_cycle, out_of_plane_selection, out_of_plane_in_or_exclude, out_of_plane_in_or_ex_range)
        
        return included_segments_oop_tilt, excluded_segments_oop_tilt_count


    def filter_segments_when_located_at_end_of_helix(self, session, alignment_size):
        helices = session.query(HelixTable).order_by(HelixTable.id).all()
        included_segments_no_ends = []
        for each_helix in helices:
            each_helix_segments = self.get_segment_ids_from_helix(session, each_helix)
            
            segments_at_helix_center = session.query(SegmentTable).\
            filter(SegmentTable.stack_id.in_(each_helix_segments)).\
            filter(and_(SegmentTable.distance_from_start_A > alignment_size/2, 
                        SegmentTable.distance_from_start_A < (each_helix.length - alignment_size/2))).all()
            
            if segments_at_helix_center is not None:
                stack_ids_at_helix_center = [ each_segment.stack_id for each_segment in segments_at_helix_center ]
                included_segments_no_ends += stack_ids_at_helix_center
            
        excluded_segments_at_ends_count = self.get_excluded_refinement_count(session, included_segments_no_ends)
        
        return included_segments_no_ends, excluded_segments_at_ends_count
    
    
    def get_selected_segments_from_last_cycle(self, ref_session):
        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
        
        selected_ref_segments = ref_session.query(RefinementCycleSegmentTable).\
        filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
        filter(RefinementCycleSegmentTable.selected == True).all()
        
        return selected_ref_segments, last_cycle
            
            
    def filter_phis_such_that_distribution_remains_even(self, phis, peaks, azimuthal_angle_count, min_bin=None):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> phis = np.array([0, 0, 0, 0, 120, 240,240,240,240,240])
        >>> s = SegmentRefine3d()
        >>> s.filter_phis_such_that_distribution_remains_even(phis, np.arange(100,110), 3)
        ([3, 2, 4, 9, 8], array([  0,   0, 120, 240, 240]))
        """
        multiple = 10.0
        if azimuthal_angle_count < multiple:
            multiple = 1
        bin_count = int(azimuthal_angle_count / float(multiple))
        freq, bound = np.histogram(phis, bin_count, (0.0, 360.0))
        mean_freq = np.mean(freq)
        cutoff_max = freq[freq <= mean_freq]
    
        if min_bin is None:
            min_bin = 2.0 * int(np.mean(cutoff_max) / float(multiple))
            
        selected_ids = []
        stack_ids = np.arange(len(phis))
        for each_id, each_bound in enumerate(bound[:-1]):
            lower_bound = each_bound
            upper_bound = bound[each_id + 1]
            filtered_peaks = peaks[(lower_bound <= phis) & (phis < upper_bound)]
            filtered_ids = stack_ids[(lower_bound <= phis) & (phis < upper_bound)]
            sorted_ids = np.flipud(np.argsort(filtered_peaks))
            sorted_filtered_ids = filtered_ids[sorted_ids]
            
            sel_id_count = min(len(sorted_filtered_ids), min_bin)
            selected_ids += sorted_filtered_ids[:sel_id_count].tolist()
        self.log.ilog('Even phi angle distribution is enforced. The following bin frequency is used as a threshold: '+ \
        '{0}'.format(min_bin))
                    
        return selected_ids, phis[selected_ids]
    

    def randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(self, pitch, rand_val, each_included_segment):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> seg_info = namedtuple('seg_info', 'inplane_angle phi helix_shift_x_A helix_shift_y_A shift_x_A shift_y_A')
        >>> each_segment = seg_info(0., 0., 0., 0., None, None)
        >>> SegmentRefine3d().randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(180, 0, each_segment)
        (0.0, 0.0, 0.0, 0.0)
        >>> SegmentRefine3d().randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(180, 1, each_segment)
        (0.0, 180.0, 0.0, 180.0)
        >>> SegmentRefine3d().randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(180, 0.5, each_segment)
        (180.0, 90.0, 0.0, 90.0)
        >>> SegmentRefine3d().randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(180, -0.5, each_segment)
        (180.0, -90.0, 0.0, -90.0)
        >>> SegmentRefine3d().randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(0, 1, each_segment)
        (0.0, 0.0, 0.0, 0.0)
        """
        phi = each_included_segment.phi + (rand_val * 360.0) % 360
        helix_shift_y_A = each_included_segment.helix_shift_y_A + rand_val * pitch
        
        x_shift, y_shift = \
        SegClassReconstruct().compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(
        each_included_segment.helix_shift_x_A, helix_shift_y_A,
        each_included_segment.inplane_angle)
        
        return phi, helix_shift_y_A, x_shift, y_shift
    
    
    def enter_results_to_segments(self, results, each_segment):
        phi, helix_shift_y_A, x_shift, y_shift = results
        
        each_segment.phi = phi
        each_segment.helix_shift_y_A = helix_shift_y_A
        each_segment.shift_x_A = x_shift
        each_segment.shift_y_A = y_shift
        
        return each_segment
    

    def enforce_even_phi_distribution(self, enforce_even_phi, release_cycle, ref_session, each_info):
        excluded_ids = []
        if enforce_even_phi:
            selected_segments, last_cycle = self.get_selected_segments_from_last_cycle(ref_session)
            
            if last_cycle.id == 1 and not self.reference_option:
                for each_included_segment in selected_segments:
                    rand_val = 2 * np.random.random() - 1
                    
                    results = self.randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(self.pitch_enforce,
                    rand_val, each_included_segment)
                    
                    each_included_segment = self.enter_results_to_segments(results, each_included_segment)
                    
                    ref_session.merge(each_included_segment)
            else:
                phis = np.array([each_segment.phi for each_segment in selected_segments])
                peaks = np.array([each_segment.peak for each_segment in selected_segments])
                sel_ids = np.array([each_segment.stack_id for each_segment in selected_segments])
                
                if self.bin_cutoff_enforce > len(phis) or self.bin_cutoff_enforce == 0:
                    min_bin = None
                else:
                    min_bin = self.bin_cutoff_enforce
                    
                filtered_ids, filtered_phis = self.filter_phis_such_that_distribution_remains_even(phis, peaks, 
                self.azimuthal_angle_count, min_bin)
                
                filt_sel_ids = sel_ids[filtered_ids]
            
                enforce_even_cycle = release_cycle / 2
                for each_selected_segment in selected_segments:
                    if each_selected_segment.stack_id not in filt_sel_ids:
                        if last_cycle.id <= enforce_even_cycle:
                            rand_val = 2 * np.random.random() - 1 
                            
                            results =\
                            self.randomize_phi_and_corresponding_helix_y_shift_based_to_pitch(self.pitch_enforce,
                            rand_val, each_selected_segment)
                            
                            each_selected_segment = self.enter_results_to_segments(results, each_selected_segment)
                            ref_session.merge(each_selected_segment)
                        elif enforce_even_cycle <= last_cycle.id <= release_cycle or min_bin is not None:
                            excluded_ids.append(each_selected_segment.stack_id)
            
            last_cycle.excluded_phi_count = len(excluded_ids)
            ref_session.commit()
        
        return excluded_ids


class SegmentRefine3dSelection(SegmentRefine3dSelectionFilter):
    
    def setup_new_refinement_db_for_each_cycle(self, ref_cycle_id):
        temp_current_ref_db = os.path.join(self.tempdir, 'ref_temp{0}{1:03}.db'.format(os.getpid(), ref_cycle_id))
        current_ref_session = SpringDataBase().setup_sqlite_db(refine_base, temp_current_ref_db)
            
        prev_ref_db = 'refinement{0:03}.db'.format(ref_cycle_id - 1)
        if os.path.exists(prev_ref_db):
            temp_prev_ref_db = self.copy_ref_db_to_tempdir(ref_cycle_id - 1)
            prev_ref_session = SpringDataBase().setup_sqlite_db(refine_base, temp_prev_ref_db)

            current_ref_session = \
            SpringDataBase().copy_all_table_data_from_one_session_to_another_session(RefinementCycleTable,
            current_ref_session, prev_ref_session)
            prev_ref_session.close()
            os.remove(temp_prev_ref_db)
                
        return current_ref_session, temp_current_ref_db
    

    def enter_refinement_parameters_in_database(self, ref_session, orientation_parameters, unbending_info,
    current_translation_step, ref_cycle_id, each_info, pixelinfo, rank=None):
        
        refinement_cycle = RefinementCycleTable()
        refinement_cycle.iteration_id = ref_cycle_id
        refinement_cycle.pixelsize = pixelinfo.pixelsize
        refinement_cycle.alignment_size_A = pixelinfo.alignment_size * pixelinfo.pixelsize
        refinement_cycle.reconstruction_size_A = pixelinfo.reconstruction_size * pixelinfo.pixelsize
        refinement_cycle.restrict_inplane = self.restrain_in_plane_rotation
        refinement_cycle.delta_inplane = self.delta_in_plane_rotation
        refinement_cycle.unbending = self.unbending
        refinement_cycle.azimuthal_restraint = each_info.azimuthal_restraint
        refinement_cycle.out_of_plane_restraint = each_info.out_of_plane_restraint
        refinement_cycle.out_of_plane_min = min(self.out_of_plane_tilt_angle_range)
        refinement_cycle.out_of_plane_max = max(self.out_of_plane_tilt_angle_range)
        refinement_cycle.out_of_plane_count = self.out_of_plane_tilt_angle_count
        refinement_cycle.azimuthal_count = self.azimuthal_angle_count
        refinement_cycle.translation_step = current_translation_step * pixelinfo.pixelsize
        refinement_cycle.x_translation_range_A = each_info.x_range * pixelinfo.pixelsize
        refinement_cycle.y_translation_range_A = each_info.y_range * pixelinfo.pixelsize
        
        for each_orient_param in orientation_parameters:
            refinement_segment = RefinementCycleSegmentTable()
            refinement_segment.cycles  = refinement_cycle
            refinement_segment.stack_id = each_orient_param.stack_id
            refinement_segment.local_id = each_orient_param.local_id
            refinement_segment.rank_id = each_orient_param.rank_id
            refinement_segment.model_id = each_orient_param.model_id
            refinement_segment.phi = each_orient_param.phi
            refinement_segment.theta = each_orient_param.theta
            refinement_segment.psi = each_orient_param.psi
            if self.unbending:
                refinement_segment.unbent_ip_angle = each_orient_param.inplane_angle
                refinement_segment.unbent_shift_x_A = each_orient_param.shift_x * pixelinfo.pixelsize
                refinement_segment.unbent_shift_y_A = each_orient_param.shift_y * pixelinfo.pixelsize
                updated_inplane_angle = -(unbending_info[each_orient_param.local_id].angle + \
                                          each_orient_param.inplane_angle) % 360
                refinement_segment.unbending_angle = updated_inplane_angle
                
                x_distance = each_orient_param.shift_x + unbending_info[each_orient_param.local_id].shift_x
                y_distance = each_orient_param.shift_y + unbending_info[each_orient_param.local_id].shift_y
                
                shift_x_A = x_distance * pixelinfo.pixelsize
                shift_y_A = y_distance * pixelinfo.pixelsize
                
                x_distance, y_distance = SegClassReconstruct().compute_distances_to_helical_axis(x_distance, y_distance,
                updated_inplane_angle)
                 
            else:
                x_distance, y_distance = \
                SegClassReconstruct().compute_distances_to_helical_axis(each_orient_param.shift_x,
                each_orient_param.shift_y, each_orient_param.inplane_angle)
                
                shift_x_A = each_orient_param.shift_x * pixelinfo.pixelsize
                shift_y_A = each_orient_param.shift_y * pixelinfo.pixelsize
                updated_inplane_angle = each_orient_param.inplane_angle
            
            refinement_segment.shift_x_A = shift_x_A
            refinement_segment.shift_y_A = shift_y_A
            refinement_segment.inplane_angle = updated_inplane_angle

#             refinement_segment.out_of_plane_angle = \
#             self.compute_out_of_plane_angle_with_respect_to_avg_inplane_angle(each_orient_param.theta, 
#             each_orient_param.psi, updated_inplane_angle)
            refinement_segment.out_of_plane_angle = each_orient_param.theta - 90.0

            refinement_segment.helix_shift_x_A = x_distance * pixelinfo.pixelsize
            refinement_segment.helix_shift_y_A = y_distance * pixelinfo.pixelsize
            
            refinement_segment.peak = each_orient_param.peak
            refinement_segment.mirror = each_orient_param.mirror
        
            ref_session.add(refinement_segment)
        
        ref_session.commit()
        
        return ref_session
    
    
    def get_exluded_ref_count_named_tuple(self):
        excluded_counts = namedtuple('refinement_counts', 'out_of_plane_tilt_count cc_prj_count helix_shift_x_count')
        return excluded_counts


    def enter_excluded_refinement_counts_in_database(self, ref_segment_count, ref_session,
    excluded_refinement_counts):
        
        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
        
        last_cycle.excluded_out_of_plane_tilt_count = excluded_refinement_counts.out_of_plane_tilt_count
        last_cycle.excluded_prj_cc_count = excluded_refinement_counts.cc_prj_count
        last_cycle.excluded_helix_shift_x_count = excluded_refinement_counts.helix_shift_x_count
        
        last_cycle.segment_count = ref_segment_count
        
        ref_session.merge(last_cycle)
        ref_session.commit()
        
    
class SegmentRefine3dParameterAveraging(SegmentRefine3dSelection):
    def get_selected_segments_from_last_refinement_cycle(self, session, ref_session, last_cycle, each_helix):
        each_helix_segments = self.get_segment_ids_from_helix(session, each_helix)
        segments = ref_session.query(RefinementCycleSegmentTable).\
        filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
        filter(RefinementCycleSegmentTable.stack_id.in_(each_helix_segments)).\
        filter(RefinementCycleSegmentTable.selected == True).all()
        
        return segments


    def get_distances_from_segment_ids(self, session, segment_ids):
        segments = session.query(SegmentTable).filter(SegmentTable.stack_id.in_(segment_ids))
        
        distances_from_start = np.array([each_segment.distances_from_start for each_segment in segments])
        
        return distances_from_start
    

    def sort_inplane_angles_into_0_360_or_180_degrees(self, selected_inplane_angles, cropped_segment_ids,
    lavg_inplane_angles, distances_from_start):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> sel = np.array([300.0, 301.0, 119.0])
        >>> ids = np.array([0, 1, 2])
        >>> angles = np.array([300.0, 300.0, 300.0])
        >>> dist = np.array([10, 20, 30])
        >>> s.sort_inplane_angles_into_0_360_or_180_degrees(sel, ids, angles, dist) #doctest: +NORMALIZE_WHITESPACE
        (array([30]), array([ 300.]), array([10, 20]), array([ 300.,  300.]), 
        array([2]), array([0, 1]), array([ 59.]), array([ 240.,  241.]))
        """
        inplane_angles_normalized = (selected_inplane_angles + lavg_inplane_angles) % 360 
        cropped_segment_ids = np.array(cropped_segment_ids)
        smaller_90_and_larger_270 = (inplane_angles_normalized < 90) ^ (inplane_angles_normalized > 270)
        close_to_0_360 = inplane_angles_normalized[smaller_90_and_larger_270]
        lavg_close_to_0_360 = lavg_inplane_angles[smaller_90_and_larger_270]
        dist_close_to_0_360 = distances_from_start[smaller_90_and_larger_270]
        ids_close_to_0_360 = cropped_segment_ids[smaller_90_and_larger_270]
        
        between_90_and_270 = (inplane_angles_normalized > 90) & (inplane_angles_normalized < 270)
        close_to_180 = inplane_angles_normalized[between_90_and_270]
        lavg_close_to_180 = lavg_inplane_angles[between_90_and_270]
        dist_close_to_180 = distances_from_start[between_90_and_270]
        ids_close_to_180 = cropped_segment_ids[between_90_and_270]
        
        return dist_close_to_0_360, lavg_close_to_0_360, dist_close_to_180, lavg_close_to_180, ids_close_to_0_360, \
        ids_close_to_180, close_to_0_360, close_to_180
    

    def compute_fitted_parameters(self, distances, parameters, new_distances=None):
        if new_distances is None:
            new_distances = distances
#        try:
#        spline_coefficients = interpolate.splrep(distances, parameters, k=2, s=3)
#        spline_fitted_parameters = interpolate.splev(new_distances, spline_coefficients)
        
        polyfit = np.polyfit(distances, parameters, max(1, int(max(distances)/1000)))
        spline_fitted_parameters = np.polyval(polyfit, new_distances)
#        except:
#            spline_fitted_parameters = parameters
        
        return spline_fitted_parameters
    

    def get_inplane_angles_per_segment_and_interpolate_two_oposite_angles(self, session, ref_session, current_cycle,
    each_helix):
        segments = self.get_selected_segments_from_last_refinement_cycle(session, ref_session, current_cycle,
        each_helix)
        
        selected_inplane_angles = np.array([each_segment.inplane_angle for each_segment in segments])
        cropped_segment_ids = [each_segment.stack_id for each_segment in segments]
        cropped_segments = session.query(SegmentTable).filter(SegmentTable.stack_id.in_(cropped_segment_ids))
        lavg_inplane_angles = np.array([each_segment.lavg_inplane_angle for each_segment in cropped_segments])
        
        if each_helix.flip_inplane_angle and current_cycle.id != 1:
            lavg_inplane_angles = (lavg_inplane_angles + 180) % 360
            flip = 1
        else:
            flip = 0
        distances_from_start = self.get_distances_from_segment_ids(session, cropped_segment_ids)
        
        dist_close_to_0_360, lavg_close_to_0_360, dist_close_to_180, lavg_close_to_180, ids_close_to_0_360, \
        ids_close_to_180, close_to_0_360, close_to_180 = \
        self.sort_inplane_angles_into_0_360_or_180_degrees(selected_inplane_angles, cropped_segment_ids,
        lavg_inplane_angles, distances_from_start)
        
        continuous_close_to_0_360 = (close_to_0_360 + 180) % 360
        spline_fitted_angles_0_360 = self.compute_fitted_parameters(dist_close_to_0_360, continuous_close_to_0_360)
        discont_spline_fitted_angles_0_360 = (spline_fitted_angles_0_360 - 180) % 360
        spline_fitted_angles_180 = self.compute_fitted_parameters(dist_close_to_180, close_to_180)
        
        return flip, segments, lavg_close_to_0_360, lavg_close_to_180, ids_close_to_0_360, ids_close_to_180, \
        close_to_0_360, close_to_180, discont_spline_fitted_angles_0_360, spline_fitted_angles_180


    def measure_inplane_angle_and_decide_for_predominant_angle(self, ref_session, segments, flip, lavg_close_to_0_360,
    lavg_close_to_180, ids_close_to_0_360, ids_close_to_180, close_to_0_360, close_to_180,
    discont_spline_fitted_angles_0_360, spline_fitted_angles_180):
        for each_ref_segment in segments:
            if each_ref_segment.stack_id in ids_close_to_0_360:
                index = ids_close_to_0_360.tolist().index(each_ref_segment.stack_id)
                lavg_inplane = (-(discont_spline_fitted_angles_0_360[index] - np.mean(lavg_close_to_0_360 + 180))) % 360
            if each_ref_segment.stack_id in ids_close_to_180:
                index = ids_close_to_180.tolist().index(each_ref_segment.stack_id)
                lavg_inplane = (-(spline_fitted_angles_180[index] - np.mean(lavg_close_to_180))) % 360
            each_ref_segment.lavg_inplane = lavg_inplane
            ref_session.add(each_ref_segment)
        
        if close_to_0_360.size >= close_to_180.size:
            avg_inplane_angle = (np.mean((close_to_0_360 + 180) % 360 - 180) - lavg_close_to_0_360) % 360
        else:
            avg_inplane_angle = (np.mean(close_to_180) - lavg_close_to_180) % 360
            flip = flip + 1
            
        return flip, avg_inplane_angle
    

    def enter_helix_inplane_parameters_in_database(self, session, ref_session, current_cycle, each_helix, flip,
    close_to_0_360, close_to_180):
        ref_helix = RefinementCycleHelixTable()
        ref_helix.segment_count_0_degree = close_to_0_360.size
        ref_helix.segment_count_180_degree = close_to_180.size
        ref_helix.cycle_id = current_cycle.id
        ref_helix.helix_id = each_helix.id
        ref_helix.flip_inplane_angle = (flip) % 2
        ref_session.add(ref_helix)
        each_helix.flip_inplane_angle = (flip) % 2
        session.merge(each_helix)


    def normalize_inplane_angles_by_picked_angles(self, picked_segment_angles, ref_inplane_angles):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> picked = np.arange(80, 90)
        >>> refined = np.arange(80.5, 90.5)
        >>> s.normalize_inplane_angles_by_picked_angles(picked, refined)
        array([ 0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5])
        """
        inplane_angles_normalized = (ref_inplane_angles - picked_segment_angles) % 360
        
        return inplane_angles_normalized
    

    def determine_predominant_side_of_angles(self, picked_segment_angles, flip, ref_inplane_angles,
    each_helix_stack_ids):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> picked = np.arange(80, 90)
        >>> flip = 0 
        >>> refined = np.arange(80.5, 90.5)
        >>> ids = np.arange(10)
        >>> s.determine_predominant_side_of_angles(picked, flip, refined, ids) #doctest: +NORMALIZE_WHITESPACE
        (0, array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]), array([], dtype=int64), 
        array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]), 
        array([ 0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5]))
        
        >>> picked = np.arange(80, 90)
        >>> flip = 0
        >>> refined = np.arange(260.5, 270.5)
        >>> ids = np.arange(20, 30)
        >>> s.determine_predominant_side_of_angles(picked, flip, refined, ids) #doctest: +NORMALIZE_WHITESPACE
        (1, array([], dtype=int64), 
        array([20, 21, 22, 23, 24, 25, 26, 27, 28, 29]), 
        array([20, 21, 22, 23, 24, 25, 26, 27, 28, 29]), 
        array([ 0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5]))
        """
        each_helix_stack_ids = np.array(each_helix_stack_ids)
        
        inplane_angles_normalized  = self.normalize_inplane_angles_by_picked_angles(picked_segment_angles,
        ref_inplane_angles)
        
        smaller_90_and_larger_270 = (inplane_angles_normalized < 90) ^ (inplane_angles_normalized > 270)
        close_to_0_360 = each_helix_stack_ids[smaller_90_and_larger_270]
        between_90_and_270 = (inplane_angles_normalized > 90) & (inplane_angles_normalized < 270)
        close_to_180 = each_helix_stack_ids[between_90_and_270]
        
        if close_to_0_360.size >= close_to_180.size:
            predominant_set = close_to_0_360
        else:
            predominant_set = close_to_180
            inplane_angles_normalized -= 180.0
            flip = flip + 1
            
        return flip, close_to_0_360, close_to_180, predominant_set, inplane_angles_normalized
    

    def get_all_selected_stack_ids(self, ref_session, current_cycle):
        selected_ref_segments = ref_session.query(RefinementCycleSegmentTable).\
        filter(RefinementCycleSegmentTable.cycle_id == current_cycle.id).\
        filter(RefinementCycleSegmentTable.selected == True).all()
        
        selected_segments = [each_ref_segment.stack_id for each_ref_segment in selected_ref_segments]
        selected_segments = list(set(selected_segments))
        
        return selected_segments


    def enter_selected_information(self, polar_helix, ref_session, ref_segments, predominant_set,
    inplane_angles_normalized, excluded_polarity_ids):
        for each_seg_index, each_ref_segment in enumerate(ref_segments):
            if each_ref_segment.stack_id in predominant_set or polar_helix in ['apolar'] or not self.force_hel_continue:
                each_ref_segment.selected = True
            else:
                each_ref_segment.selected = False
                excluded_polarity_ids.append(each_ref_segment.stack_id)
            each_ref_segment.norm_inplane_angle = inplane_angles_normalized[each_seg_index]
            ref_session.merge(each_ref_segment)
        
        return excluded_polarity_ids
    

    def exclude_inplane_angles_outside_delta_psi(self, restrain_in_plane_rotation, delta_in_plane_rotation, ref_session,
    current_cycle):
        if restrain_in_plane_rotation:
            selected_ref_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == current_cycle.id).\
            filter(RefinementCycleSegmentTable.selected == True).\
            filter(or_(*
                       [and_(*[RefinementCycleSegmentTable.norm_inplane_angle > delta_in_plane_rotation, 
                               RefinementCycleSegmentTable.norm_inplane_angle < 180.0 - delta_in_plane_rotation]), 
                        and_(*[RefinementCycleSegmentTable.norm_inplane_angle > 180.0 + delta_in_plane_rotation, 
                               RefinementCycleSegmentTable.norm_inplane_angle < 360.0 - delta_in_plane_rotation])]
                       )).all()
                       
            for each_selected_segment in selected_ref_segments:
                each_selected_segment.selected = False
                ref_session.merge(each_selected_segment)
            
            excluded_inplane_ids = [each_selected_segment.stack_id for each_selected_segment in selected_ref_segments]
            ref_session.commit()
        else:
            excluded_inplane_ids = []
            
        return excluded_inplane_ids
    

    def select_segments_based_on_in_plane_rotation(self, session, ref_session, last_cycle, helices, polar_helix,
    restrain_in_plane_rotation, delta_in_plane_rotation, included_non_orientation):
        
        excluded_polarity_ids = []
        for each_helix in helices:
            each_helix_segments = session.query(SegmentTable).filter(SegmentTable.helix_id == each_helix.id).all()

            each_helix_segment_ids = [each_segment.stack_id for each_segment in each_helix_segments 
                                      if each_segment.stack_id in included_non_orientation]
            picked_segment_angles = np.array([each_segment.lavg_inplane_angle for each_segment in each_helix_segments
                                      if each_segment.stack_id in included_non_orientation])
            
            if each_helix.flip_inplane_angle:
                picked_segment_angles = (picked_segment_angles + 180) % 360
                flip = 1
            else:
                flip = 0 
            
            ref_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
            filter(RefinementCycleSegmentTable.stack_id.in_(each_helix_segment_ids)).\
            order_by(RefinementCycleSegmentTable.id).all()
            
            ref_inplane_angles = np.array([each_ref_segment.inplane_angle for each_ref_segment in ref_segments])
            
            flip, close_to_0_360, close_to_180, predominant_set, inplane_angles_normalized = \
            self.determine_predominant_side_of_angles(picked_segment_angles, flip, ref_inplane_angles,
            each_helix_segment_ids)
                
            self.enter_helix_inplane_parameters_in_database(session, ref_session, last_cycle, each_helix, flip,
            close_to_0_360, close_to_180)
            
            excluded_polarity_ids = self.enter_selected_information(polar_helix, ref_session, ref_segments,
            predominant_set, inplane_angles_normalized, excluded_polarity_ids)
            
        ref_session.commit()
        
        excluded_inplane_ids = self.exclude_inplane_angles_outside_delta_psi(restrain_in_plane_rotation,
        delta_in_plane_rotation, ref_session, last_cycle)
            
        last_cycle.excluded_inplane_count = len(set(excluded_polarity_ids + excluded_inplane_ids))
        ref_session.commit()
        session.commit()
            
        selected_segments = self.get_all_selected_stack_ids(ref_session, last_cycle)
        
        return selected_segments
        
        
#     def compute_out_of_plane_angle_with_respect_to_inplane_angle(self, theta, inplane_angle):
#         if 90.0 < inplane_angle <= 270.0:
#             out_of_plane_angle = -(90.0 - theta)
#         else:
#             out_of_plane_angle = 90 - theta
#             
#         return out_of_plane_angle
#     
#             
#     def compute_out_of_plane_angle_with_respect_to_avg_inplane_angle(self, theta, psi, avg_inplane_angle):
#         """
#         >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
#         >>> s = SegmentRefine3d()
#         >>> s.compute_out_of_plane_angle_with_respect_to_avg_inplane_angle(90, 275, 5)
#         0
#         >>> s.compute_out_of_plane_angle_with_respect_to_avg_inplane_angle(86, 275, 5)
#         4
#         >>> s.compute_out_of_plane_angle_with_respect_to_avg_inplane_angle(86, 95, 5)
#         -4
#         >>> s.compute_out_of_plane_angle_with_respect_to_avg_inplane_angle(94, 95, 5)
#         4
#         >>> s.compute_out_of_plane_angle_with_respect_to_avg_inplane_angle(94, 275, 5)
#         -4
#         """
#         psi_norm = (psi - avg_inplane_angle - 270.0) % 360
#         if psi_norm < 90 or psi_norm > 270:
#             out_of_plane_angle = 90 - theta
#         elif 90 <= psi_norm <= 270:
#             out_of_plane_angle = -(90 - theta)
#         
#         return out_of_plane_angle
        

    def get_all_distances_and_selection_mask_from_ref_segments(self, session, ref_session, last_cycle, each_helix,
    included_non_orientation):
        
        each_helix_segments = session.query(SegmentTable).filter(SegmentTable.helix_id == each_helix.id).all()
        
        all_helix_segment_ids = [each_segment.stack_id for each_segment in each_helix_segments
                                 if each_segment.stack_id in included_non_orientation]
        all_distances_from_start = [each_segment.distance_from_start_A for each_segment in each_helix_segments
                                 if each_segment.stack_id in included_non_orientation]
        
        if all_helix_segment_ids != []:
            all_ref_helix_segments = self.get_all_segments_from_refinement_cycle(ref_session, last_cycle,
            all_helix_segment_ids)
        
            excluded_segments = np.invert([bool(each_segment.selected) for each_segment in all_ref_helix_segments])
        else:
            all_ref_helix_segments = []
            excluded_segments = []

        all_shift_y = np.array([each_segment.helix_shift_y_A for each_segment in all_ref_helix_segments])
        all_distances_from_start += all_shift_y
            
        return all_ref_helix_segments, each_helix_segments, excluded_segments, all_distances_from_start
    

    def compute_fit_if_more_than_three_datapoints(self, all_distances_from_start, quantity, sel_distances_from_start,
    selected_quantity):
        if len(sel_distances_from_start) > 3:
            spline_fitted_inplane_angle = self.compute_fitted_parameters(sel_distances_from_start, selected_quantity,
            all_distances_from_start)
        else:
            spline_fitted_inplane_angle = quantity

        return spline_fitted_inplane_angle


    def update_average_in_plane_rotation_angle_per_helix(self, session, ref_session, last_cycle, helices,
    included_non_orientation):
    
        for each_helix in helices:
            all_ref_helix_segments, each_helix_segments, excluded_segments, all_distances_from_start = \
            self.get_all_distances_and_selection_mask_from_ref_segments(session, ref_session, last_cycle, each_helix,
            included_non_orientation)
            
            if all_ref_helix_segments != []:
                inplane_angles_normalized = np.array([each_ref_segment.norm_inplane_angle for each_ref_segment in \
                all_ref_helix_segments])
                
                if self.polar_helix in ['polar']:
                    inplane_angles_normalized = (inplane_angles_normalized + 180.0) % 360
                elif self.polar_helix in ['apolar']:
                    inplane_angles_normalized = (inplane_angles_normalized + 90.0) % 180
                
                sel_distances_from_start = np.ma.masked_array(all_distances_from_start, mask=excluded_segments).compressed()
                selected_inplane_angles = np.ma.masked_array(inplane_angles_normalized, mask=excluded_segments).compressed()
                
                spline_fitted_inplane_angle = self.compute_fit_if_more_than_three_datapoints(all_distances_from_start,
                inplane_angles_normalized, sel_distances_from_start, selected_inplane_angles)
                
                if self.polar_helix in ['polar']:
                    spline_fitted_inplane_angle = (spline_fitted_inplane_angle - 180.0 ) % 360
                elif self.polar_helix in ['apolar']:
                    spline_fitted_inplane_angle = (spline_fitted_inplane_angle - 90.0 ) % 360
                
                picked_segment_angles = np.array([each_segment.lavg_inplane_angle for each_segment in each_helix_segments
                                                  if each_segment.stack_id in included_non_orientation])
                
                if each_helix.flip_inplane_angle:
                    picked_segment_angles = (picked_segment_angles + 180.0) % 360
                    
                corrected_inplane_angles = self.normalize_inplane_angles_by_picked_angles(-picked_segment_angles,
                spline_fitted_inplane_angle)
                
                for each_index, each_ref_segment in enumerate(all_ref_helix_segments):
                    each_ref_segment.lavg_inplane_angle = corrected_inplane_angles[each_index]
                    ref_session.merge(each_ref_segment)
            
        ref_session.commit()
        
    
    def update_average_out_of_plane_per_helix(self, session, ref_session, last_cycle, helices,
    included_non_orientation):
        
        for each_helix in helices:
            all_ref_helix_segments, each_helix_segments, excluded_segments, all_distances_from_start = \
            self.get_all_distances_and_selection_mask_from_ref_segments(session, ref_session, last_cycle, each_helix,
            included_non_orientation)
            
            if all_ref_helix_segments != []:
                out_of_plane_angles = [each_segment.out_of_plane_angle for each_segment in all_ref_helix_segments]
    
                selected_out_of_plane = np.ma.masked_array(out_of_plane_angles, mask=excluded_segments).compressed()
                sel_distances_from_start = np.ma.masked_array(all_distances_from_start, mask=excluded_segments).compressed()
                
                spline_fitted_outofplane = self.compute_fit_if_more_than_three_datapoints(all_distances_from_start,
                out_of_plane_angles, sel_distances_from_start, selected_out_of_plane)
                
                for each_index, each_ref_segment in enumerate(all_ref_helix_segments):
                    each_ref_segment.lavg_out_of_plane = spline_fitted_outofplane[each_index]
                    ref_session.merge(each_ref_segment)
            
        ref_session.commit()
        
    
    def update_average_helix_shift_x_per_helix(self, session, ref_session, last_cycle, helices,
    included_segments_non_orientation):
        for each_helix in helices:
            all_ref_helix_segments, each_helix_segments, excluded_segments, all_distances_from_start = \
            self.get_all_distances_and_selection_mask_from_ref_segments(session, ref_session, last_cycle, each_helix,
            included_segments_non_orientation)
            
            if all_ref_helix_segments != []:
                all_shift_x = np.array([each_segment.helix_shift_x_A for each_segment in all_ref_helix_segments])
                
                sel_distances_from_start = np.ma.masked_array(all_distances_from_start, mask=excluded_segments).compressed()
                selected_shift_x = np.ma.masked_array(all_shift_x, mask=excluded_segments).compressed()
                
                spline_fitted_shift_x = self.compute_fit_if_more_than_three_datapoints(all_distances_from_start,
                all_shift_x, sel_distances_from_start, selected_shift_x)
                
                for each_index, each_ref_segment in enumerate(all_ref_helix_segments):
                    each_ref_segment.lavg_helix_shift_x_A = spline_fitted_shift_x[each_index]
                    ref_session.merge(each_ref_segment)
            
        ref_session.commit()
        
            
    def enter_final_ids_into_database(self, ref_session, last_cycle, final_combined_ids):
        self.log.tlog('The following segments are included in the reconstruction ' + \
        'procedure:\n{0}'.format(', '.join([str(each_seg_id) for each_seg_id in final_combined_ids])))
        
        ref_segments = ref_session.query(RefinementCycleSegmentTable).\
        filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).all()
        
        total_excluded = 0
        for each_segment in ref_segments:
            if not self.force_hel_continue:
                each_segment.out_of_plane_angle = each_segment.theta - 90.0
            if each_segment.stack_id in final_combined_ids:
                each_segment.selected = True
            else:
                each_segment.selected = False
                total_excluded += 1
            ref_session.merge(each_segment)
        
        last_cycle.total_excluded_count = total_excluded 
        
        ref_session.merge(last_cycle)
        ref_session.commit()
    

    def determine_forward_difference_and_set_ref_segments(self, ref_segments, helix_shift_x_A, attr_to_set):
        diffs = np.diff(helix_shift_x_A)
        forward_diffs = (np.append(diffs, 0) + np.insert(diffs, 0, 0)) / np.sqrt(2.0)

        [setattr(each_ref_segment, attr_to_set, forward_diffs[each_id]) 
         for each_id, each_ref_segment in enumerate(ref_segments)]
        
        return forward_diffs, ref_segments
        

    def compute_forward_difference_for_selected_segments_and_select(self, session, ref_session, last_cycle):
        
        helices = session.query(HelixTable).order_by(HelixTable.id).all()
        
        peaks = np.array([])
        excluded_helix_shift_x_count = 0
        included_segments_helix_shift_x = np.array([])
        for each_helix in helices:
            each_helix_segments = session.query(SegmentTable).filter(SegmentTable.helix_id == each_helix.id).all()
            each_helix_segment_ids = [each_helix_segment.stack_id for each_helix_segment in each_helix_segments]
            
            ref_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
            filter(RefinementCycleSegmentTable.selected == True).\
            filter(RefinementCycleSegmentTable.stack_id.in_(each_helix_segment_ids)).all()

            helix_shift_x_A = np.array([each_ref_segment.helix_shift_x_A for each_ref_segment in ref_segments])
            inplane_rot = (np.array([each_ref_segment.norm_inplane_angle for each_ref_segment in ref_segments]) + 90) % 180
            outofplane_tilt = np.abs([each_ref_segment.out_of_plane_angle for each_ref_segment in ref_segments])

            if 'apolar' in self.polar_helices:
                helix_shift_x_A = np.abs(helix_shift_x_A)

            forward_diffs, ref_segments = self.determine_forward_difference_and_set_ref_segments(ref_segments,
            helix_shift_x_A, 'forward_diff_x_shift_A')
            
            forward_diffs_rot, ref_segments = self.determine_forward_difference_and_set_ref_segments(ref_segments,
            inplane_rot, 'forward_diff_inplane')
            
            forward_diffs_tilt, ref_segments = self.determine_forward_difference_and_set_ref_segments(ref_segments,
            outofplane_tilt, 'forward_diff_outofplane')
            
            abs_forward_diffs = np.abs(forward_diffs)
            if self.helix_shift_x_selection and self.helix_shift_x_in_or_exclude == 'include':
                ex_seg = [setattr(each_ref_segment, 'selected', False) for each_id, each_ref_segment in enumerate(ref_segments)\
                 if abs_forward_diffs[each_id] > self.helix_shift_x_in_or_ex_cutoff]

                incl_ref_stack_ids = [each_ref_segment.stack_id for each_id, each_ref_segment in enumerate(ref_segments)\
                 if abs_forward_diffs[each_id] <= self.helix_shift_x_in_or_ex_cutoff]
            elif self.helix_shift_x_selection and self.helix_shift_x_in_or_exclude == 'exclude':
                ex_seg = [setattr(each_ref_segment, 'selected', False) for each_id, each_ref_segment in enumerate(ref_segments)\
                 if abs_forward_diffs[each_id] <= self.helix_shift_x_in_or_ex_cutoff]

                incl_ref_stack_ids = [each_ref_segment.stack_id for each_id, each_ref_segment in enumerate(ref_segments)\
                 if abs_forward_diffs[each_id] > self.helix_shift_x_in_or_ex_cutoff]
            else:
                ex_seg = []
                incl_ref_stack_ids = np.array([each_ref_segment.stack_id for each_ref_segment in ref_segments])
            
            excluded_helix_shift_x_count += len(ex_seg)
            included_segments_helix_shift_x = np.append(included_segments_helix_shift_x, incl_ref_stack_ids)

            [ref_session.merge(each_ref_segment) for each_id, each_ref_segment in enumerate(ref_segments)]
            
            peaks = np.append(peaks, [each_ref_segment.peak for each_ref_segment in ref_segments])
        
        last_cycle.mean_peak = np.mean(peaks)
        
        ref_session.merge(last_cycle)
        ref_session.commit()
        
        included_segments_helix_shift_x = np.int32(included_segments_helix_shift_x)
        included_segments_helix_shift_x = included_segments_helix_shift_x.tolist()

        return included_segments_helix_shift_x, excluded_helix_shift_x_count
            

    def prepare_databases_for_selection(self, orientation_parameters, unbending_info, current_translation_step,
    ref_cycle_id, each_info, pixelinfo):
        self.log.fcttolog()
        self.log.in_progress_log()
        
        mean_peaks = [each_param.peak for each_param in orientation_parameters]

        ccc_proj_range_vals = SegmentSelect().convert_relative_range_to_absolute_range_values(mean_peaks,
        self.ccc_proj_range)

        ref_session, temp_current_ref_db = self.setup_new_refinement_db_for_each_cycle(ref_cycle_id)
        ref_session = self.enter_refinement_parameters_in_database(ref_session, orientation_parameters, 
        unbending_info, current_translation_step, ref_cycle_id, each_info, pixelinfo)
        
        return ref_session, temp_current_ref_db, ccc_proj_range_vals
    

    def select_segments_based_on_out_of_plane_tilt(self, ref_session, session, last_cycle):
        included_segments_oop_tilt, excluded_oop_tilt_count = self.filter_segments_by_out_of_plane_tilt(session,
        ref_session, last_cycle, self.out_of_plane_selection, self.out_of_plane_in_or_exclude,
        self.out_of_plane_in_or_ex_range)

        ref_segments = ref_session.query(RefinementCycleSegmentTable).\
        filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
        filter(RefinementCycleSegmentTable.selected == True).all()
        
        for each_sel_segment in ref_segments:
            if each_sel_segment.stack_id not in included_segments_oop_tilt:
                each_sel_segment.selected = False
                ref_session.merge(each_sel_segment)
        
        ref_session.commit()
        
        return included_segments_oop_tilt, excluded_oop_tilt_count


    def select_refinement_parameters_based_on_selection_criteria_hierarchically(self, ref_session, ccc_proj_range_vals,
    session, last_cycle, helices, included_segments_non_orientation):
        ref_segments = ref_session.query(RefinementCycleSegmentTable).all()

        ref_stack_ids = [each_ref_segment.stack_id for each_ref_segment in ref_segments]
        ref_segment_count = len(ref_stack_ids)

        included_segments_non_orientation = list(set(included_segments_non_orientation).intersection(ref_stack_ids))

        included_segments_inplane = self.select_segments_based_on_in_plane_rotation(session, ref_session, last_cycle, 
        helices, self.polar_helix, self.restrain_in_plane_rotation, self.delta_in_plane_rotation, 
        included_segments_non_orientation)

        included_segments_oop_tilt, excluded_oop_tilt_count = \
        self.select_segments_based_on_out_of_plane_tilt(ref_session, session, last_cycle)

        included_segments_helix_shift_x, excluded_helix_shift_x_count = \
        self.compute_forward_difference_for_selected_segments_and_select(session, ref_session, last_cycle)

        included_segments_ccc_proj, excluded_cc_prj_count = self.filter_segments_by_ccc_against_projections(session, 
        ref_session, last_cycle, self.ccc_proj_selection, self.ccc_proj_in_or_exclude, ccc_proj_range_vals)

        excluded_counts = self.get_exluded_ref_count_named_tuple()

        excluded_refinement_counts = excluded_counts(excluded_oop_tilt_count, excluded_cc_prj_count,
        excluded_helix_shift_x_count)

        included_ref_segments = set(included_segments_oop_tilt).intersection(included_segments_ccc_proj,
        included_segments_helix_shift_x, included_segments_inplane)

        return included_ref_segments, ref_segment_count, excluded_refinement_counts


    def perform_helix_based_computations_and_selection(self, each_info, spring_db, ref_session, ccc_proj_range_vals):
        session = SpringDataBase().setup_sqlite_db(base, spring_db)
        
        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
        helices = session.query(HelixTable).order_by(HelixTable.id).all()
        
        included_segments_non_orientation, excluded_non_orientation_counts = \
        SegmentSelect().filter_non_orientation_parameters_based_on_selection_criteria(self, spring_db,
        keep_helices_together=True)
        
        included_ref_segments, ref_segment_count, excluded_refinement_counts =\
        self.select_refinement_parameters_based_on_selection_criteria_hierarchically(ref_session, ccc_proj_range_vals,
        session, last_cycle, helices, included_segments_non_orientation)

        if self.force_hel_continue:
            self.update_average_in_plane_rotation_angle_per_helix(session, ref_session, last_cycle, helices,
            included_segments_non_orientation)
            
            self.update_average_out_of_plane_per_helix(session, ref_session, last_cycle, helices,
            included_segments_non_orientation)
            
            self.update_average_helix_shift_x_per_helix(session, ref_session, last_cycle, helices,
            included_segments_non_orientation)
            
        excluded_phi_ids = self.enforce_even_phi_distribution(self.enforce_even_phi, self.release_cycle, ref_session,
        each_info)
        
        included_segments_non_orientation, excluded_non_orientation_counts = \
        SegmentSelect().filter_non_orientation_parameters_based_on_selection_criteria(self, spring_db,
        keep_helices_together=False)
        
        self.enter_excluded_refinement_counts_in_database(ref_segment_count, ref_session,
        excluded_refinement_counts)
            
        final_combined_ids = list(set(included_ref_segments).intersection(included_segments_non_orientation).\
                                  difference(excluded_phi_ids))
        
        self.enter_final_ids_into_database(ref_session, last_cycle, final_combined_ids)
        
        return ref_session
    

    def get_helices_from_corresponding_frames(self, session, ref_session):
        first_mic = session.query(CtfMicrographTable).first()

        frame_ending = first_mic.micrograph_name.split('@')[-1]

        unique_mics = session.query(CtfMicrographTable).\
        filter(CtfMicrographTable.micrograph_name.endswith(frame_ending)).all()
        unique_mic_ids = [each_mic.id for each_mic in unique_mics]
            
        helices = session.query(HelixTable).order_by(HelixTable.id).all()
        unique_helices = [each_helix for each_helix in helices if each_helix.mic_id in unique_mic_ids]
        helix_ids = []
        for each_helix in unique_helices:
            segment = session.query(SegmentTable).filter(SegmentTable.helix_id == each_helix.id).first()
            segment_mic = session.query(CtfMicrographTable).filter(CtfMicrographTable.id == segment.mic_id).first()

            frame_prefix = segment_mic.micrograph_name.split('@')[0]

            frame_mics = session.query(CtfMicrographTable).\
            filter(CtfMicrographTable.micrograph_name.startswith(frame_prefix)).all()
            
            frame_mic_ids = [each_frame_mic.id for each_frame_mic in frame_mics]

            frame_segments = session.query(SegmentTable).\
            filter(SegmentTable.picked_x_coordinate_A == segment.picked_x_coordinate_A).\
            filter(SegmentTable.mic_id.in_(frame_mic_ids)).order_by(SegmentTable.helix_id).all()
            
            helix_ids.append([each_frame_segment.helix_id for each_frame_segment in frame_segments])

        return helix_ids


    def average_shifts_between_frames_along_helix(self, x_shifts, window_size):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> shifts = [[each_frame] * 20 for each_frame in list(range(10))]
        >>> shifts = [list(range(each_frame, each_frame + 12)) for each_frame in list(range(10))]
        >>> shifts = np.array(shifts) ** 2
        >>> shifts
        array([[  0,   1,   4,   9,  16,  25,  36,  49,  64,  81, 100, 121],
               [  1,   4,   9,  16,  25,  36,  49,  64,  81, 100, 121, 144],
               [  4,   9,  16,  25,  36,  49,  64,  81, 100, 121, 144, 169],
               [  9,  16,  25,  36,  49,  64,  81, 100, 121, 144, 169, 196],
               [ 16,  25,  36,  49,  64,  81, 100, 121, 144, 169, 196, 225],
               [ 25,  36,  49,  64,  81, 100, 121, 144, 169, 196, 225, 256],
               [ 36,  49,  64,  81, 100, 121, 144, 169, 196, 225, 256, 289],
               [ 49,  64,  81, 100, 121, 144, 169, 196, 225, 256, 289, 324],
               [ 64,  81, 100, 121, 144, 169, 196, 225, 256, 289, 324, 361],
               [ 81, 100, 121, 144, 169, 196, 225, 256, 289, 324, 361, 400]])
        >>> avg_shifts = s.average_shifts_between_frames_along_helix(shifts.tolist(), 3)
        >>> np.int64(avg_shifts)
        array([[ -6,   1,   4,   9,  16,  25,  36,  48,  63,  80,  99, 123],
               [ -3,   4,   9,  16,  25,  36,  49,  63,  80,  99, 120, 146],
               [  0,   9,  16,  25,  36,  49,  63,  80,  99, 120, 143, 170],
               [  6,  16,  25,  36,  49,  64,  81,  99, 121, 143, 168, 196],
               [ 15,  25,  36,  49,  64,  81, 100, 121, 144, 169, 196, 225],
               [ 25,  36,  48,  63,  81, 100, 121, 144, 169, 196, 225, 255],
               [ 38,  48,  63,  80,  99, 120, 144, 169, 196, 225, 256, 288],
               [ 52,  63,  80,  99, 120, 143, 169, 196, 225, 256, 289, 322],
               [ 68,  80,  99, 120, 143, 168, 196, 225, 256, 289, 324, 358],
               [ 87,  99, 120, 143, 168, 195, 225, 256, 289, 324, 361, 397]])
        """
        x_shifts = np.vstack(x_shifts)

        helix_length = len(x_shifts[0])
        avg_x_shifts = np.copy(x_shifts)
        for each_segment in list(range(helix_length)):
            avg_x_shifts[:,each_segment]=np.average(x_shifts[:,each_segment])
        
        x_shift_diffs = x_shifts - avg_x_shifts

        avg_x_shift_diffs = np.zeros(x_shifts.shape)
        for each_id, each_frame in enumerate(x_shift_diffs):
            avg_x_shift_diffs[each_id] = SegmentSelect().compute_local_average_from_measurements(each_frame, window_size)
        
        helix_avg_x_shifts = avg_x_shifts + avg_x_shift_diffs

        return helix_avg_x_shifts


    def perform_local_averaging_across_frames(self, session, ref_session, helix_ids):
        if len(helix_ids) > 0:
            segments = session.query(SegmentTable).filter(SegmentTable.helix_id == helix_ids[0][0]).all()
            distance = np.average(np.diff([each_segment.distance_from_start_A for each_segment in segments]))
            window_size = int(round(self.frame_local_avg_dstnce / distance))

            ref_segments = ref_session.query(RefinementCycleSegmentTable).all()

            for each_helix_ids in helix_ids:
                x_shifts = []
                y_shifts = []
                stack_ids = []
                for each_helix_id in each_helix_ids:
                    segments = session.query(SegmentTable).filter(SegmentTable.helix_id == each_helix_id).all()
                    hel_stack_ids = [each_segment.stack_id for each_segment in segments]
                    
                    ref_helix_segments = [each_ref_segment for each_ref_segment in ref_segments \
                                          if each_ref_segment.stack_id in hel_stack_ids]
                    
                    stack_ids.append(hel_stack_ids)
                    x_shifts.append([each_ref_segment.shift_x_A for each_ref_segment in ref_helix_segments])
                    y_shifts.append([each_ref_segment.shift_y_A for each_ref_segment in ref_helix_segments])
            
                stack_ids = np.vstack(stack_ids)
    
                x_shifts = self.average_shifts_between_frames_along_helix(x_shifts, window_size)
                y_shifts = self.average_shifts_between_frames_along_helix(y_shifts, window_size)
            
            stack_ids = stack_ids.ravel()
            x_shifts = x_shifts.ravel()
            y_shifts = y_shifts.ravel()
        else:
            stack_ids = []
            x_shifts = []
            y_shifts = []

        return stack_ids, x_shifts, y_shifts


    def sort_and_enter_averaged_shifts(self, ref_session, shift_info):
        stack_iids, x_shifts, y_shifts = shift_info

        argsort_ids = np.argsort(stack_iids)
        shift_info = zip(stack_iids[argsort_ids], x_shifts[argsort_ids], y_shifts[argsort_ids])

        for each_stack_id, each_x_shift, each_y_shift in shift_info:
            ref_segment = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.stack_id == int(each_stack_id)).first()

            ref_segment.shift_x_A = each_x_shift
            ref_segment.shift_y_A = each_y_shift
            ref_session.merge(ref_segment)
        
        ref_session.commit()


    def perform_local_frame_averaging_and_ref_database_update(self, spring_db, ref_db, ref_session):
        if self.frame_motion_corr and self.frame_local_avg_dstnce > 0:
            session = SpringDataBase().setup_sqlite_db(base, spring_db)
            helix_ids = self.get_helices_from_corresponding_frames(session, ref_session)
            shift_info = self.perform_local_averaging_across_frames(session, ref_session, helix_ids)
            self.sort_and_enter_averaged_shifts(ref_session, shift_info)


    def get_selected_alignment_parameters_from_last_cycle(self, ref_cycle_id, model_id, rank):
        temp_ref_db = self.copy_ref_db_to_tempdir(ref_cycle_id)
        ref_session = SpringDataBase().setup_sqlite_db(refine_base, temp_ref_db)
        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()
        if rank is None:
            selected_ref_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
            filter(RefinementCycleSegmentTable.model_id == model_id).\
            filter(RefinementCycleSegmentTable.selected == True).all()
        elif rank is not None:
            selected_ref_segments = ref_session.query(RefinementCycleSegmentTable).\
            filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
            filter(RefinementCycleSegmentTable.model_id == model_id).\
            filter(RefinementCycleSegmentTable.rank_id == rank).\
            filter(RefinementCycleSegmentTable.selected == True).all()
            
        ref_session.close()
        os.remove(temp_ref_db)
        
        return selected_ref_segments
    
    
    def update_total_nonorientation_counts_in_ref_db(self, ref_cycle_id, spring_db, ref_session):
        included_segments_non_orientation, excluded_non_orientation_counts = \
        SegmentSelect().filter_non_orientation_parameters_based_on_selection_criteria(self, spring_db,
        keep_helices_together=False)
        
        last_cycle = ref_session.query(RefinementCycleTable).order_by(desc(RefinementCycleTable.id)).first()

        last_cycle.excluded_mic_count = excluded_non_orientation_counts.mic_count
        last_cycle.excluded_helix_count = excluded_non_orientation_counts.helix_count
        last_cycle.excluded_class_count = excluded_non_orientation_counts.class_count
        last_cycle.excluded_curvature_count = excluded_non_orientation_counts.curvature_count
        last_cycle.excluded_defocus_count = excluded_non_orientation_counts.defocus_count
        last_cycle.excluded_astigmatism_count = excluded_non_orientation_counts.astig_count
        last_cycle.excluded_layer_cc_count = excluded_non_orientation_counts.layer_cc_count
        last_cycle.excluded_helix_ends_count = 0

        ref_session.merge(last_cycle)
        ref_session.commit()
        ref_session.close()


    def prepare_refined_alignment_parameters_from_database(self, ref_cycle_id, pixelsize, unbending, reference_files,
    rank=None):
        comb_orientation_parameters = []
        for each_reference in reference_files:
            selected_ref_segments = self.get_selected_alignment_parameters_from_last_cycle(ref_cycle_id,
            each_reference.model_id, rank)
            
            orientation_parameters = []
            rec_parameters = SegClassReconstruct().make_named_tuple_for_reconstruction()
            for each_segment in selected_ref_segments:
                if unbending:
                    orientation_parameters.append(rec_parameters(each_segment.stack_id, each_segment.local_id,
                    each_segment.phi, each_segment.theta, each_segment.psi, each_segment.unbent_shift_x_A / pixelsize,
                    each_segment.unbent_shift_y_A / pixelsize, each_segment.unbent_ip_angle, each_segment.mirror,
                    each_segment.id))
                else:
                    orientation_parameters.append(rec_parameters(each_segment.stack_id, each_segment.local_id,
                    each_segment.phi, each_segment.theta, each_segment.psi, each_segment.shift_x_A / pixelsize,
                    each_segment.shift_y_A / pixelsize, each_segment.inplane_angle, each_segment.mirror, each_segment.id))
            
            comb_orientation_parameters.append(orientation_parameters)
        
        return comb_orientation_parameters
        
    
    def select_segments_based_on_specified_criteria(self, orientation_parameters, unbending_info,
    current_translation_step, ref_cycle_id, each_info, pixelinfo, reference_files):
    
        ref_session, temp_ref_db, ccc_proj_range_vals = self.prepare_databases_for_selection(orientation_parameters,
        unbending_info, current_translation_step, ref_cycle_id, each_info, pixelinfo)
        
        temp_db = self.copy_spring_db_to_tempdir()
        self.perform_local_frame_averaging_and_ref_database_update(temp_db, temp_ref_db, ref_session)
            
        ref_session = self.perform_helix_based_computations_and_selection(each_info, temp_db, ref_session,
        ccc_proj_range_vals)
        
        self.update_total_nonorientation_counts_in_ref_db(ref_cycle_id, temp_db, ref_session)

        os.remove(temp_db)
        shutil.copy(temp_ref_db, 'refinement{0:03}.db'.format(ref_cycle_id))
        os.remove(temp_ref_db)
        
        selected_parameters = self.prepare_refined_alignment_parameters_from_database(ref_cycle_id, pixelinfo.pixelsize,
        self.unbending, reference_files)
        
        return selected_parameters


#!/usr/bin/env python
"""
Test module to check michelixtracegrid
"""
from glob import glob
from multiprocessing import cpu_count
import os
import shutil
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.tests.csinfrastr.test_csreadinput import EndToEndTest
from spring.tests.micprgs.test_michelixtrace import TestMicHelixTrace
from spring.micprgs.michelixtracegrid import MicHelixTraceGrid, MicHelixTraceGridPar


class TestMicHelixTraceGridSetup(TestMicHelixTrace):
    def setup_of_common_mhtg_non_mpi_parameters(self):
        self.feature_set = MicHelixTraceGridPar()
#         self.input_stack = 'test_mic.hdf'
        self.prim_var = 'alpha_threshold'
        self.sec_var = 'min_helix_length'
        self.grid_db = os.path.join(os.pardir, 'grid.db')
        self.prim_range = ((1e-7, 1e-3))
        self.sec_range = ((300.0, 700.0))
        self.prim_inc = 0.0001998
        self.sec_inc = 100.0
        
        self.feature_set.parameters['First parameter']=self.prim_var
        self.feature_set.parameters['Second parameter']=self.sec_var
        self.feature_set.parameters['Lower and upper limit first parameter']=self.prim_range
        self.feature_set.parameters['Lower and upper limit second parameter']=self.sec_range
        self.feature_set.parameters['First and second parameter increment']=((self.prim_inc, self.sec_inc))
        self.feature_set.parameters['Subgrid option']=False
        self.feature_set.parameters['Part and number of subgrids']=(3, 3)
        self.feature_set.parameters['Grid continue option']=False
        self.feature_set.parameters['Grid database']=self.grid_db
        self.setup_main_michelixtrace_parameters()
        

class TestMicHelixTraceGrid(TestMicHelixTraceGridSetup, MicHelixTraceGrid):
    def setup_mhtg(self):
        self.setup_of_common_mhtg_non_mpi_parameters()
        
        self.feature_set.parameters['Compute performance score']=True
        self.feature_set.parameters['MPI option']=False
        self.feature_set.parameters['Number of CPUs']=min(3, cpu_count())
        self.feature_set.parameters['Temporary directory']=os.path.abspath(os.curdir)
        
        super(TestMicHelixTraceGridSetup, self).__init__(self.feature_set)
        self.write_groud_truth_coord_file()
        
    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()
        
        self.setup_mhtg()

    def teardown(self):
        pass
        os.remove(self.helix_box_file)
         
        parfiles = glob('*par')
        logfiles = glob('*log')
        directory = 'alpha_threshold'
        dirfiles = glob('{dir}*{sep}*'.format(dir=directory, sep=os.sep))
        for each_dir in dirfiles:
            os.chdir(os.path.dirname(each_dir))
            self.remove_mic_dir_including_box_files() 
            os.chdir(os.pardir)
         
        for each_file in logfiles + parfiles + dirfiles:
            if os.path.isfile(each_file):
                os.remove(each_file)
            elif os.path.isdir(each_file):
                os.rmdir(each_file)
 
        for each_dir in glob('{0}*'.format(directory)):
            shutil.rmtree(each_dir)
            
        for each_file in glob('gri*.db'):
            os.remove(each_file)
          
        self.remove_mic_dir_including_box_files()
 
        self.testingdir.remove()
            
        
class TestMicHelixTraceGridMain(TestMicHelixTraceGrid):
    def do_test_case_mhtg1(self):
        """
        Test standard Michelixtracegrid
        """
        self.launch_michelixtrace_jobs()


class TestMicHelixTraceGridMpi(TestMicHelixTraceGrid):
    def do_end_to_end_test_mhtg_inputfile(self):
        """
        Test standard Michelixtracegrid MPI option
        """
        self.feature_set.parameters['MPI option']=True
        self.feature_set.parameters['Number of CPUs']=min(3, cpu_count())

        EndToEndTest().do_end_to_end_inputfile(self.feature_set)

 
class TestMicHelixTraceGridEndToEnd(TestMicHelixTraceGrid):
    def do_end_to_end_test_mhtg_inputfile(self):
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)
 

    def do_end_to_end_test_mhtg_prompt(self):
        EndToEndTest().do_end_to_end_prompt(self.feature_set)
        
        
def main():
    tmhtg = TestMicHelixTraceGridMain()
    tmhtg.setup()
    tmhtg.do_test_case_mhtg1()
        
if __name__ == '__main__':
    main()



# Author: Carsten Sachse 17-Feb-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Spring (Single Particle Reconstruction from Images of kNown Geometries) - 
suite of programs for processing cryo-EM images of helices
"""
from PyQt4.QtCore import SIGNAL, SLOT, Qt, QRect
from PyQt4.QtGui import QMessageBox, QAction, QIcon, QMenu, QMenuBar, QToolBar, \
    QWidget
from spring.csinfrastr.cslogger import GetMetaData
from spring.csinfrastr.csproductivity import ExtLauncher
import os


class GuiActionsMenuTrigger(object):

    def openManual(self):
        """openManual(self)
        Opens the manual supplied with this application in the user's
        web browser.
        """
        web_address = 'http://www.sachse.embl.de/emspring'
        
        self.manual_browser = ExtLauncher()
        self.manual_browser.qlaunch_open_file(web_address)
        
        
class GuiActionsMenu(GuiActionsMenuTrigger):
    def createFileMenu(self):
        # open
        
        self.openAction = QAction(QIcon(os.path.join(os.path.dirname(__file__),
        '{pardir}{sep}images{sep}icons{sep}open.png'.format(pardir=os.pardir, sep=os.sep))), 'Open Parameter File',
        self)
        
        self.openAction.setShortcut('Ctrl+O')
        self.openAction.setStatusTip('Open')

        self.saveAction = QAction(QIcon(os.path.join(os.path.dirname(__file__),
        '{pardir}{sep}images{sep}icons{sep}save.png'.format(pardir=os.pardir, sep=os.sep))), 'Save Parameter File',
        self)
        
        self.saveAction.setShortcut('Ctrl+S')
        self.saveAction.setStatusTip('Save')

        # exit
        
        self.exitAction = QAction(QIcon(os.path.join(os.path.dirname(__file__),
        '{pardir}{sep}images{sep}icons{sep}exit.png'.format(pardir=os.pardir, sep=os.sep))), 'Exit Spring', self)
        
        self.exitAction.setShortcut('Ctrl+Q')
        self.exitAction.setStatusTip('Exit')
        self.connect(self.exitAction, SIGNAL('triggered()'), SLOT('close()'))

        # File menu
        self.menu_File = QMenu(self.menubar)
        self.menu_File.setTitle('&File')
        self.menu_File.setObjectName('menu_File')
        self.menu_File.addAction(self.openAction)
        self.menu_File.addAction(self.saveAction)
        self.menu_File.addAction(self.exitAction)


    def createViewMenu(self):
        self.beginnerAction = QAction('Beginner', self)
        self.beginnerAction.setShortcut('Ctrl+1')
        self.beginnerAction.setStatusTip('Beginner')
        self.beginnerAction.setToolTip('Displays beginner-level spring programs')

        self.intermediateAction = QAction('Intermediate', self)
        self.intermediateAction.setShortcut('Ctrl+2')
        self.intermediateAction.setStatusTip('Intermediate')
        self.intermediateAction.setToolTip('Displays intermediate-level spring programs')

        self.expertAction = QAction('Expert', self)
        self.expertAction.setShortcut('Ctrl+3')
        self.expertAction.setStatusTip('Expert')
        self.expertAction.setToolTip('Displays expert-level spring programs')

        #  View menu
        self.menu_View = QMenu(self.menubar)
        self.menu_View.setTitle('&View')
        self.menu_View.setObjectName('menu_View')
        self.menu_View.addAction(self.beginnerAction)
        self.menu_View.addAction(self.intermediateAction)
        self.menu_View.addAction(self.expertAction)
        

    def createHelpMenu(self):
        ## create actions
        # about
        self.aboutAction = QAction('Information about Spring', self)
        self.aboutAction.setStatusTip('About')
        self.connect(self.aboutAction, SIGNAL('triggered()'), self.helpAbout)

        # manual
        
        self.manualAction = QAction(QIcon(os.path.join(os.path.dirname(__file__),
        '{pardir}{sep}images{sep}icons{sep}help.png'.format(pardir=os.pardir, sep=os.sep))), 'Documentation - Open ' + \
        'Manual', self)
        
        self.manualAction.setStatusTip('Documentation')
        self.connect(self.manualAction, SIGNAL('triggered()'), self.openManual)

        # Help menu
        self.menu_Help = QMenu(self.menubar)
        self.menu_Help.setTitle('&Help')
        self.menu_Help.setObjectName('menu_Help')
        self.menu_Help.addAction(self.aboutAction)
        self.menu_Help.addAction(self.manualAction)


class GuiActions(QWidget, GuiActionsMenu):
    def setupUi(self, Window):

        self.statusBar()
        self.menubar = QMenuBar(self)
        self.menubar.setGeometry(QRect(0,0,543,27))
        self.setMenuBar(self.menubar)

        self.createFileMenu()
        self.createViewMenu()
        self.createHelpMenu()
        
        self.menubar.addAction(self.menu_File.menuAction())
        self.menubar.addAction(self.menu_View.menuAction())
        self.menubar.addAction(self.menu_Help.menuAction())

        ## Toolbar
        self.toolbar = QToolBar()
#        self.toolbar.setGeometry(QRect(100,100,543,27))
#        self.toolbar.setGeometry(QRect(100,500,543,27))
#        self.toolbar.setOrientation(Qt.Vertical)
        self.toolbar.addAction(self.exitAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.openAction)
        self.toolbar.addAction(self.saveAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.beginnerAction)
        self.toolbar.addAction(self.intermediateAction)
        self.toolbar.addAction(self.expertAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.manualAction)
        
        self.addToolBar(Qt.ToolBarArea(Qt.LeftToolBarArea), self.toolbar)

    def helpAbout(self): 
        meta = GetMetaData(self.package)
        eggmeta = meta.release_meta()
        QMessageBox.about(self, 'About Scambl',
            """<b>%s (%s)</b> <p>%s <p>&copy; %s <p>Author: %s <p>Maintainer: %s, %s <p>License: %s<p>Status: %s""" % \
            (self.package.title(), eggmeta['Version'], __doc__, eggmeta['Copyright'], eggmeta['Author'],
            eggmeta['Maintainer'], eggmeta['Author-email'], eggmeta['License'], 
            eggmeta['Classifier: Development Status']))




# Author: Carsten Sachse 
# with Stefan Huber (2017) 
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
from EMAN2 import EMNumPy
import matplotlib
from matplotlib.mlab import griddata
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np

from spring.csinfrastr.csproductivity import DiagnosticPlot
from scipy import interpolate, ndimage, optimize, stats
from collections import namedtuple


class MicHelixTraceSupport(object):
    def roundint(self, arr):
            return np.rint(arr).astype(np.int)
            
            
    def interpolate_chain(self, chain, step_A=1):
        lenx, leny = np.abs(chain[-1] - chain[0])
        x, y = chain[:, 0], chain[:, 1]
        if lenx >= leny:
            f = interpolate.interp1d(x, y)
            x_new = np.arange(x.min()+step_A, x.max(), step_A)
            return np.vstack((x_new, f(x_new))).T
        else:
            f = interpolate.interp1d(y, x)
            y_new = np.arange(y.min()+step_A, y.max(), step_A)
            return np.vstack((f(y_new), y_new)).T
    
    
    def setup_boolean_grid(self, bundle, sizex_A, sizey_A, gridsize_A, chain_interpolation_A=1):
        gridsizex, gridsizey = sizex_A // gridsize_A + 1, sizey_A // gridsize_A + 1
        grid = np.zeros((gridsizey, gridsizex))
        for chain in bundle:
            chain = self.roundint(self.interpolate_chain(chain, chain_interpolation_A) / float(gridsize_A))
            chain = chain[chain[:, 1] < gridsizex]
            chain = chain[chain[:, 0] < gridsizey]
            grid[chain[:, 1], chain[:, 0]] = 1

        return grid
    
    
    def clean_boolean_grid(self, boolean_grid, gridsize_A, edge_clean=400, helix_ends_clean=True,
    helix_ends_clean_radius=100, helix_bundle=None, minimum_distance_helixends=150):
        """
        :param boolean_grid:
        :param gridsize_A:
        :param edge_clean: Are edges of the micrograph cleaned.
        :param helix_ends_clean: Clean helix ends out of boolean grid
        :param helix_ends_clean_radius: What radius to use for cleaning
        :param helix_bundle: Bundle of ground truth helices from helixboxer, used to find helix ends
        :param minimum_distance_helixends: Helix ends are just removed if above a minimum distance.
        Prevents discontinous helices from being removed.
        :return: copy of boolean_grid
        """
    
        boolean_grid = np.copy(boolean_grid)
    
        edge_clean_pixel = edge_clean // gridsize_A
    
    
        boolean_grid[0:edge_clean_pixel] = False
        boolean_grid[-edge_clean_pixel:] = False
        boolean_grid[:, 0:edge_clean_pixel] = False
        boolean_grid[:, -edge_clean_pixel:] = False
    
        if helix_ends_clean:
            helix_ends = np.vstack(([[i[0], i[-1]] for i in helix_bundle]))
            from scipy.spatial.distance import pdist, squareform
            helix_ends_distant = helix_ends[(squareform(pdist(helix_ends)) < minimum_distance_helixends).sum(0) == 1]
            for cx,cy in helix_ends_distant:
                cx, cy = self.roundint(cx/float(gridsize_A)), self.roundint(cy/float(gridsize_A))
                radius = self.roundint(helix_ends_clean_radius/float(gridsize_A))
                index = self.prepare_circular_mask(radius)
                clean_area = boolean_grid[cy - radius:cy + radius + 1, cx - radius:cx + radius + 1]
                if clean_area.shape == index.shape:
                    boolean_grid[cy - radius:cy + radius + 1, cx - radius:cx + radius + 1][index] = False
    
        return boolean_grid
    
    
    def prepare_circular_mask(self, size):
        X, Y = [np.arange(-size, size + 1)] * 2
        disk_mask = X[:, None] ** 2 + Y ** 2 <= size ** 2

        return disk_mask
    
    
    def inflate_binary(self, binary, size):
        disk_mask = self.prepare_circular_mask(size)
        binary_inflated = ndimage.binary_dilation(binary, structure=disk_mask).astype(binary.dtype)

        return binary_inflated
    

    def compute_precision_and_recall(self, grid_truth, grid_truth_inflate, grid_test, grid_test_inflate):
        """
        Recall is also called TPR or Sensitivity. Likelihood that a true helix is really found as a positive test outcome
        Precision is also called positive predictive value, Likelihood that positive test outcome is correct
        """
        P = grid_truth.sum()
        Test_Outcome_Positive = grid_test.sum()
    
        TP_recall = (grid_test_inflate * grid_truth).sum()  # Number of squares found when inflating result
        TP_precision = (grid_test * grid_truth_inflate).sum()  # Number of correctly found squares when inflating truth
    
        return [TP_recall, TP_precision, P, Test_Outcome_Positive]
        
    
    def compare_interactively_traced_with_ground_truth(self, helices_ground_truth, helices_traced, size_y, size_x,
    ori_pixelsize, helixwidth_A):
        sizex_A, sizey_A = self.roundint(np.array([size_x, size_y]) * ori_pixelsize)
        gridsize_A = 25
        radius_A = helixwidth_A / 2.0
        
        michelix = self.setup_boolean_grid(helices_traced, sizex_A, sizey_A, gridsize_A)
        michelix_inflate = self.inflate_binary(michelix, radius_A / gridsize_A)
    
        michelix_inflate_clean = self.clean_boolean_grid(michelix_inflate, gridsize_A, edge_clean=400, helix_ends_clean=True,
                                      helix_ends_clean_radius=radius_A*2,
                                      helix_bundle=helices_ground_truth)
    
        michelix_clean = self.clean_boolean_grid(michelix, gridsize_A, edge_clean=400, helix_ends_clean=True,
                                          helix_ends_clean_radius=radius_A*2,
                                          helix_bundle=helices_ground_truth)
    
        helixboxer = self.setup_boolean_grid(helices_ground_truth, sizex_A, sizey_A, gridsize_A)
        helixboxer_inflate = self.inflate_binary(helixboxer, radius_A / gridsize_A)
    
        helixboxer_inflate_clean = self.clean_boolean_grid(helixboxer_inflate, gridsize_A, edge_clean=400, helix_ends_clean=True,
                                        helix_ends_clean_radius=radius_A*2,
                                        helix_bundle=helices_ground_truth)
    
        helixboxer_clean = self.clean_boolean_grid(helixboxer, gridsize_A, edge_clean=400, helix_ends_clean=True,
                                        helix_ends_clean_radius=radius_A*2,
                                        helix_bundle=helices_ground_truth)
                #[TP_recall, TP_precision, P, Test_Outcome_Positive]
        P_R = self.compute_precision_and_recall(helixboxer_clean, helixboxer_inflate_clean, michelix_clean,
        michelix_inflate_clean)
    
        return P_R
                                        
    
    def summarize_parameter_info_over_micrographs_old(self, tracing_results_mic):
        """
        >>> from spring.micprgs.michelixtrace_helperfunctions import MicHelixTraceSupport
        >>> params = [['test_mic.hdf', 200, 0.5, 0.0, 0.0, 37.0, 157.0], ['test_mic.hdf', 400, 0.5, 0.0, 0.0, 37.0, 71.0], ['test_mic.hdf', 600, 0.5, 0.0, 0.0, 37.0, 71.0], ['test_mic.hdf', 900, 0.5, 0.0, 0.0, 37.0, 0.0], ['test_mic.hdf', 200, 0.01, 0.0, 0.0, 37.0, 317.0], ['test_mic.hdf', 400, 0.01, 0.0, 0.0, 37.0, 334.0], ['test_mic.hdf', 600, 0.01, 0.0, 0.0, 37.0, 281.0], ['test_mic.hdf', 900, 0.01, 0.0, 0.0, 37.0, 196.0], ['test_mic.hdf', 200, 0.001, 0.0, 0.0, 37.0, 293.0], ['test_mic.hdf', 400, 0.001, 0.0, 0.0, 37.0, 313.0], ['test_mic.hdf', 600, 0.001, 0.0, 0.0, 37.0, 290.0], ['test_mic.hdf', 900, 0.001, 0.0, 0.0, 37.0, 227.0], ['test_mic.hdf', 200, 1e-05, 0.0, 0.0, 37.0, 212.0], ['test_mic.hdf', 400, 1e-05, 0.0, 0.0, 37.0, 226.0], ['test_mic.hdf', 600, 1e-05, 0.0, 0.0, 37.0, 174.0], ['test_mic.hdf', 900, 1e-05, 0.0, 0.0, 37.0, 38.0]]
        >>> MicHelixTraceSupport().summarize_parameter_info_over_micrographs_old(params)
        array([[  2.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00],
               [  2.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00],
               [  2.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00],
               [  2.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                  0.00000000e+00],
               [  4.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00],
               [  4.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00],
               [  4.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00],
               [  4.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                  0.00000000e+00],
               [  6.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00],
               [  6.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00],
               [  6.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00],
               [  6.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                  0.00000000e+00],
               [  9.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00],
               [  9.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00],
               [  9.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00],
               [  9.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                             nan]])
    
        """
        # Content is: [micrograph_name, each_min_helix_length, alpha each_threshold, TP_recall, TP_precision, P, Test_outcome_positive]
        tracing_results_mic = np.stack(tracing_results_mic)
    
        tracing_results_mic[:,0] = 0
        tracing_results_mic = tracing_results_mic.astype(np.float)
        min_helix_lengths = np.unique(tracing_results_mic[:,1])
        thresholds = np.unique(tracing_results_mic[:,2])
        tracing_criteria_comb = []
        for each_min_helix_length in min_helix_lengths:
            for each_threshold in thresholds:
                summary_all_micrographs = tracing_results_mic[
                    (tracing_results_mic[:,1]==each_min_helix_length) & 
                    (tracing_results_mic[:,2]==each_threshold)].sum(0)
                    
                summary_recall = summary_all_micrographs[3] / summary_all_micrographs[5]
                summary_precision = summary_all_micrographs[4] / summary_all_micrographs[6]
                tracing_criteria_comb.append([each_min_helix_length, each_threshold, summary_recall, summary_precision])
    
        return np.array(tracing_criteria_comb) #[each_min_helix_length, each_threshold, summary_recall, summary_precision]
        
        
    def summarize_parameter_info_over_micrographs(self, tracing_results_mic):
        """
        >>> from spring.micprgs.michelixtrace_helperfunctions import MicHelixTraceSupport
        >>> params = [['test_mic.hdf', 200, 0.5, 0.0, 0.0, 37.0, 157.0], ['test_mic.hdf', 400, 0.5, 0.0, 0.0, 37.0, 71.0], ['test_mic.hdf', 600, 0.5, 0.0, 0.0, 37.0, 71.0], ['test_mic.hdf', 900, 0.5, 0.0, 0.0, 37.0, 0.0], ['test_mic.hdf', 200, 0.01, 0.0, 0.0, 37.0, 317.0], ['test_mic.hdf', 400, 0.01, 0.0, 0.0, 37.0, 334.0], ['test_mic.hdf', 600, 0.01, 0.0, 0.0, 37.0, 281.0], ['test_mic.hdf', 900, 0.01, 0.0, 0.0, 37.0, 196.0], ['test_mic.hdf', 200, 0.001, 0.0, 0.0, 37.0, 293.0], ['test_mic.hdf', 400, 0.001, 0.0, 0.0, 37.0, 313.0], ['test_mic.hdf', 600, 0.001, 0.0, 0.0, 37.0, 290.0], ['test_mic.hdf', 900, 0.001, 0.0, 0.0, 37.0, 227.0], ['test_mic.hdf', 200, 1e-05, 0.0, 0.0, 37.0, 212.0], ['test_mic.hdf', 400, 1e-05, 0.0, 0.0, 37.0, 226.0], ['test_mic.hdf', 600, 1e-05, 0.0, 0.0, 37.0, 174.0], ['test_mic.hdf', 900, 1e-05, 0.0, 0.0, 37.0, 38.0]]
        >>> crit = MicHelixTraceSupport().summarize_parameter_info_over_micrographs(params)
        >>> np.array(crit)
        array([[  2.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  2.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  2.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  2.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  4.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  4.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  4.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  4.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  6.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  6.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  6.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  6.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  9.00000000e+02,   1.00000000e-05,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  9.00000000e+02,   1.00000000e-03,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  9.00000000e+02,   1.00000000e-02,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00],
               [  9.00000000e+02,   5.00000000e-01,   0.00000000e+00,
                  0.00000000e+00,   0.00000000e+00,   0.00000000e+00]])

        """
        # Content is: 
        #[micrograph_name, each_min_helix_length, alpha each_threshold, TP_recall, TP_precision, P, Test_outcome_positive]
        tracing_results_mic = np.stack(tracing_results_mic)
    
        tracing_results_mic[:,0] = 0
        tracing_results_mic = tracing_results_mic.astype(np.float)
        min_helix_lengths = np.unique(tracing_results_mic[:,1])
        thresholds = np.unique(tracing_results_mic[:,2])
        tracing_criteria_comb = []
        criteria_tuple = namedtuple('criteria', 'min_length threshold recall precision f1_measure f05_measure')
        for each_min_helix_length in min_helix_lengths:
            for each_threshold in thresholds:
                summary_all_micrographs = tracing_results_mic[
                    (tracing_results_mic[:,1]==each_min_helix_length) & 
                    (tracing_results_mic[:,2]==each_threshold)].sum(0)
                    
                if summary_all_micrographs[5] > 0:
                    recall = summary_all_micrographs[3] / summary_all_micrographs[5]
                else:
                    recall = 0.0

                if summary_all_micrographs[6] > 0:
                    precision = summary_all_micrographs[4] / summary_all_micrographs[6]
                else:
                    precision = 0.0

                f1_score = self.compute_fscore_b(1 / 1.0, precision, recall)
                f05_score = self.compute_fscore_b(1 / 2.0, precision, recall)
                tracing_criteria_comb.append(
                    criteria_tuple(each_min_helix_length, each_threshold, recall, precision, f1_score, f05_score))
    
        return tracing_criteria_comb
        
        
    def interpolate_parameter_space(self, trcng_crit_comb):
        alphas = np.array([each_trcng_result.threshold for each_trcng_result in trcng_crit_comb])
        mincut = np.array([each_trcng_result.min_length for each_trcng_result in trcng_crit_comb])

        x = np.unique(alphas)
        y = np.unique(mincut)
    
        recalls = np.array([each_trcng_result.recall for each_trcng_result in trcng_crit_comb])
        precisions = np.array([each_trcng_result.precision for each_trcng_result in trcng_crit_comb])
        
        xi = np.logspace(np.log10(x.min()), np.log10(x.max()), 70)
        yi = np.linspace(y.min(), y.max(), 70)
        zi_precisions = griddata(alphas, mincut, precisions, xi, yi, interp='linear')
        zi_recalls = griddata(alphas, mincut, recalls, xi, yi, interp='linear')
    
        return xi, yi, zi_precisions, zi_recalls
    
        
    def compute_fscore_b_old(self, beta, prec, recall):
        fscore = (1+beta**2) * prec * recall / (beta**2*prec+recall)
        fscore[np.isinf(fscore)] = 0
    
        return np.nan_to_num(fscore)


    def compute_fscore_b(self, beta, prec, recall):
        """
        >>> from spring.micprgs.michelixtrace_helperfunctions import MicHelixTraceSupport
        >>> MicHelixTraceSupport().compute_fscore_b(1/2.0, 1, 0)
        0.0
        >>> MicHelixTraceSupport().compute_fscore_b(1/2.0, 0, 1)
        0.0
        >>> MicHelixTraceSupport().compute_fscore_b(1/2.0, 1, 1)
        1.0
        >>> MicHelixTraceSupport().compute_fscore_b(1/2.0, 0.25, 0.75)
        0.28846153846153844
        >>> MicHelixTraceSupport().compute_fscore_b(1/2.0, 0, 0)
        0.0
        >>> a = np.zeros(10)
        >>> a[0] = 1.0
        >>> MicHelixTraceSupport().compute_fscore_b(1/2.0, a, a)
        array([ 1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.])
        """
        if np.isscalar(prec) and prec + recall == 0:
            fscore = 0.0
        else:
            fscore = (1  + beta**2) * prec * recall / (beta**2 * prec + recall)

        if not np.isscalar(fscore):
            fscore = np.nan_to_num(fscore)
    
        return fscore

    
    def plot_parameter_search_summary(self, xi, yi, zi_precisions, zi_recalls, absolutethresholdoption):
        matplotlib.rcParams.update({'font.size': 10})
        contour_fontsize = 8
    
        fig, [[ax1,ax2],[ax3,ax4]] = plt.subplots(2,2, figsize=(7,7))
    
        def plot_parameter_contour(ax, xi, yi, zi, cont_font, title):
            ax.pcolormesh(xi, yi, zi, cmap = plt.get_cmap('rainbow'))
            CS = ax.contour(xi, yi, zi, colors = 'k')
            ax.clabel(CS, inline=1, fontsize=cont_font)
            ax.set_title(title)
            ax.set_xscale("log") 
            if absolutethresholdoption:
                ax.set_xlabel('Absolute CC threshold')
            else:
                ax.set_xlabel('Alpha threshold (Adaptive)')
            ax.set_ylabel('Minimum Helix Cutoff')
    
        plot_parameter_contour(ax1, xi, yi, zi_precisions, contour_fontsize, 'Precision')
        plot_parameter_contour(ax2, xi, yi, zi_recalls, contour_fontsize, 'Recall')        
    
        beta = 1/2.0
        fscore = self.compute_fscore_b(beta, zi_precisions, zi_recalls)
        plot_parameter_contour(ax3, xi, yi, fscore.data, contour_fontsize, 'F%.2f Measure'%(beta)) 
        maxidx = np.unravel_index(fscore.argmax(), fscore.shape)
        max_x, max_y = xi[maxidx[1]], yi[maxidx[0]]
        prec, rec = zi_precisions[maxidx[0], maxidx[1]], zi_recalls[maxidx[0], maxidx[1]]
        ax3.scatter(max_x, max_y, color='white', edgecolor='k', s=50, 
                    label='%.3g, %.0f\nPrecision %.2g\nRecall %.2g'%(max_x, max_y, prec, rec))
        ax3.legend(fontsize=7)
    
        beta = 1/1.0
        fscore = self.compute_fscore_b(beta, zi_precisions, zi_recalls)
        plot_parameter_contour(ax4, xi, yi, fscore.data, contour_fontsize, 'F%.2f Measure'%(beta)) 
        maxidx = np.unravel_index(fscore.argmax(), fscore.shape)
        max_x, max_y = xi[maxidx[1]], yi[maxidx[0]]
        prec, rec = zi_precisions[maxidx[0], maxidx[1]], zi_recalls[maxidx[0], maxidx[1]]
        ax4.scatter(max_x, max_y, color='white', edgecolor='k', s=50, 
                    label='%.3g, %.0f\nPrecision %.2g\nRecall %.2g'%(max_x, max_y, prec, rec))
        ax4.legend()
        ax4.legend(fontsize=7)
        
        for ax in [ax1,ax2,ax3,ax4]:
            ax.set_xlim(xi.min(), xi.max())
            ax.set_ylim(yi.min(), yi.max())
        
        fig.tight_layout()
        fig.savefig('ParameterSpace.pdf')
        
        return max_x, max_y #pair of (best threshold, best min_helix_length) is returned
    
    
    def compute_persistence_length_from_tangent_vector_correlation(self, bundle):
        """Bundle is a list of 'chains' - shape (n,2) arrays, where n can differ between the arrays.
           Length of segments is allowed to be different within a chain and between chains"""
        correlations = []
        distances = []
        for chain in bundle:
            tangent_vecs = chain[1:] - chain[:-1]
            unit_tangent_vecs = tangent_vecs / np.sqrt(np.sum(tangent_vecs ** 2, 1))[:, None]
            segment_dist_matrix = self.compute_segment_dist_matrix(chain)
            correlation = np.inner(unit_tangent_vecs, unit_tangent_vecs)
            correlations.extend(correlation.flatten().tolist())
            distances.extend(segment_dist_matrix.flatten().tolist())
    
        correlations = np.array(correlations)
        distances = np.array(distances)
    
        bins = np.linspace(distances[distances!=0].min()*2.0, np.percentile(distances, 70), 20)
    
        bin_means, bin_edges, binnumber = stats.binned_statistic(distances, correlations, statistic = 'mean',
                                                                 bins = bins)
        bin_stds,  bin_edges, binnumber = stats.binned_statistic(distances, correlations, statistic = 'std',
                                                                 bins = bins)
        bin_stds /= np.sqrt(np.bincount(binnumber)[0:len(bin_stds)])
        bin_width = (bin_edges[1] - bin_edges[0])
        bin_centers = bin_edges[1:] - bin_width / 2
    
        def expon(reclamb, x):
            return np.exp(-reclamb*x) #reclamb symbolises 1/lamb
    
        # Interpolate possible nan values
        nan_values = np.isnan(bin_means)
        bin_means[nan_values] = np.interp(np.flatnonzero(nan_values), np.flatnonzero(~nan_values), bin_means[~nan_values])

        # Curve fit
        reclamb, dreclamb = optimize.curve_fit(expon, bin_centers, bin_means, p0=0.001, sigma=None, absolute_sigma=False)
        lamb = 1/reclamb[0]
    
        # Error propagation, relative error of reclam is also relative error of lamb
        dlamb = np.sqrt(dreclamb[0,0])/reclamb[0]*lamb
    
        return distances, correlations, bin_centers, bin_means, bin_stds, lamb, dlamb
    
    
    def compute_segment_dist_matrix(self, chain):
        n_segments = len(chain) - 1
    
        segment_mids = (chain[1:] + chain[:-1]) / 2.0
        segment_mids_vectors = segment_mids[1:] - segment_mids[:-1]
        segment_distances = np.sqrt(np.sum(segment_mids_vectors ** 2, 1))
        segment_distances_cumulated = np.hstack((0, np.cumsum(segment_distances)))
    
        segment_dist_matrix = np.zeros((n_segments, n_segments))
        for i in range(n_segments):
            jump_distances = (segment_distances_cumulated[i + 1:] - segment_distances_cumulated[0:-i - 1])
            rgn = np.arange(len(jump_distances))
            segment_dist_matrix[rgn, rgn + 1 + i] = jump_distances
    
        segment_dist_matrix += segment_dist_matrix.T
    
        return segment_dist_matrix
    
    
    def compute_mad(self, data):
        return 1.4826 * np.mean(np.absolute(data - np.mean(data)))
    
    def plot_pers_length_summary(self, bundle, pl_exact, dpl_exact, pl_list, distances, correlations, bin_centers,
                                 bin_means, bin_stds, pruning_cutoff, plotfile):
        pl_list = pl_list[~np.isnan(pl_list)]
        fig, axarr = plt.subplots(2, 2, figsize=(11.69, 8.27))
    
        ex = [0, np.percentile(distances, 95), np.percentile(correlations, 3),  1]
        hexplot = axarr[1, 0].hexbin(distances, correlations, gridsize=20, cmap='gist_heat_r', extent=ex,
            norm=matplotlib.colors.LogNorm()) #, bins='log'
        axarr[1, 0].scatter(bin_centers, bin_means, s=5, color='green', label='Binned mean')
        axarr[1, 0].legend(fontsize=12)
        axarr[1, 0].set_xlabel('Polymer Length [Angstrom]')
        axarr[1, 0].set_ylabel('dot product of tangent vectors')
        axarr[1, 0].set_xlim(ex[0:2])
        axarr[1, 0].set_ylim(ex[2:])
        plt.colorbar(hexplot,ax=axarr[1, 0], label='Histogram Count')
        axarr[1, 0].get_yaxis().get_major_formatter().set_useOffset(False)
    
        bins = 10 ** np.linspace(np.log10(pl_list.min() / 10000.0), np.log10(pl_list.max() / 10000.0), 50)
        n, bins, patches = axarr[0, 1].hist(pl_list / 10000.0, color='darkolivegreen', bins=bins)
        axarr[0, 1].set_xlabel('Persistence Length [um]')
        axarr[0, 1].set_ylabel('# of helices')
    
        median = np.median(pl_list)
        axarr[0, 1].axvline(median/10000.0, color='blue', label='Distribution median = %.3g um' % (median/10000.0))
        axarr[0, 1].axvline(np.exp(np.log(median) + self.compute_mad(np.log(pl_list)))/10000.0, color='blue', linestyle='dotted',
                            label='Median Absolute Deviation ~ 1 StdDev')
        axarr[0, 1].axvline(np.exp(np.log(median) - self.compute_mad(np.log(pl_list)))/10000.0, color='blue', linestyle='dotted')
        pruning_cutoff_absolute = np.exp(np.log(median) - float(pruning_cutoff)*self.compute_mad(np.log(pl_list)))/ 10000.0
        axarr[0, 1].axvline(pruning_cutoff_absolute, color='orangered',
                            linestyle='-.', label='Persistence Length Cutoff %.3g um'%pruning_cutoff_absolute)
        axarr[0, 1].legend(fontsize=9)
        axarr[0, 1].set_title('Persistence Length Measure per Helix')
        axarr[0, 1].set_xscale("log")
        for b, p in zip(bins, patches):
            if b < pruning_cutoff_absolute:
                plt.setp(p, 'facecolor', 'red')
    
    #     bin_width = bin_centers[1] - bin_centers[0]
    #     bin_left = bin_centers - bin_width/2.0
        bin_centers_zero = np.hstack(([0], bin_centers))
        bin_means_zero = np.hstack(([1], bin_means))
        bin_stds_zero = np.hstack(([0], bin_stds))
        np.savetxt('TangentVectorDotProduct.txt', np.column_stack((bin_centers_zero, bin_means_zero, bin_stds_zero)),
                   header = 'Polymer Length\tCorrelation\tCorr Std\tPL=%.5f'%pl_exact, delimiter='\t')
        axarr[1, 1].errorbar(bin_centers_zero, bin_means_zero, label='Real Data\nbinned mean and se', color='darkolivegreen', yerr=bin_stds_zero)
        axarr[1, 1].plot(bin_centers_zero, np.exp(-1 / pl_exact * bin_centers_zero),
                         label='Exponential Fit\nLambda = \n(%.3g +- %.2g) um' % (pl_exact / 10000.0, dpl_exact / 10000.0),
                         color='black', linestyle='dotted')
        axarr[1, 1].set_ylim(ex[2:])
        axarr[1, 1].set_xlabel('Polymer Length [Angstrom]')
        axarr[1, 1].set_ylabel('dot product of tangent vectors')
        axarr[1, 1].legend(fontsize=12, loc=3)
        axarr[1, 1].get_yaxis().get_major_formatter().set_useOffset(False)
    
        colormap = plt.get_cmap('autumn')
        colormap.set_under(color='black')
        vmax, vmin = pl_list.max()/10000.0, pruning_cutoff_absolute
        norm = matplotlib.colors.LogNorm(vmin, vmax)
        sm = plt.cm.ScalarMappable(cmap=colormap, norm=norm)
        sm._A = []
        if len(bundle) > 300:
            for i in np.random.choice(len(bundle), 300):
                lineplot = axarr[0, 0].plot(*bundle[i].T, color=colormap(norm(pl_list[i]/10000)))
        else:
            for i, chain in enumerate(bundle):
                lineplot = axarr[0, 0].plot(*chain.T, color=colormap(norm(pl_list[i]/10000)))
        axarr[0, 0].set_xlabel('x [Angstrom]')
        axarr[0, 0].set_ylabel('y [Angstrom]')
        axarr[0, 0].set_title('300 random helix traces (black: too curved)')
        axarr[0, 0].set_aspect('equal', adjustable='box')
        axarr[0, 0].tick_params(axis='both', which='major', labelsize=8)
        if vmax>vmin:
            plt.colorbar(sm, ax=axarr[0, 0], label='Persistence Length Measure [um]')
    
        fig.tight_layout()
        fig.savefig(plotfile)
    
        return pruning_cutoff_absolute
    
    
    def remove_ticks_and_scale_correctly(self, ax, mic_np):
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlim([0, mic_np.shape[1]])
        ax.set_ylim([0, mic_np.shape[0]])
    
        return ax
    
    
    def visualize_traces_in_diagnostic_plot_verbose(self, infile, outfile, overlap_cc, binary, helices, mic, ref,
                                                    cross_corr, rho, angles, peaks, xy_center_grid, fitfunct,
                                                    fitparams, skel, pers_lengths, feature_set, a_threshold):
        mic.read_image(infile)
        mic_np = np.copy(EMNumPy.em2numpy(mic))
    
        michelixtrace_plot = DiagnosticPlot()
        fig = michelixtrace_plot.add_header_and_footer(feature_set, infile, outfile)
    
        ax1 = michelixtrace_plot.plt.subplot2grid((2, 5), (0, 0), colspan=1, rowspan=1)  # Overl_CC
        ax2 = michelixtrace_plot.plt.subplot2grid((2, 5), (1, 0), colspan=1, rowspan=1)  # Binary CC
        ax2_hist = michelixtrace_plot.plt.subplot2grid((2, 5), (0, 1), colspan=1, rowspan=1)  # Overl_CC Hist
        ax2_hist_trans = michelixtrace_plot.plt.subplot2grid((2, 5), (1, 1), colspan=1, rowspan=1)  # Overl_CC Hist
        ax3 = michelixtrace_plot.plt.subplot2grid((2, 5), (0, 2), colspan=1, rowspan=1)  # Micrograph with helices
        ax4 = michelixtrace_plot.plt.subplot2grid((2, 5), (1, 2), colspan=1, rowspan=1)  # Reference Helix
        ax5 = michelixtrace_plot.plt.subplot2grid((2, 5), (0, 3), colspan=1, rowspan=1)  # 2DCC
        ax6 = michelixtrace_plot.plt.subplot2grid((2, 5), (1, 3), colspan=1, rowspan=1)  # Reconstruction Vectors
        ax7 = michelixtrace_plot.plt.subplot2grid((2, 5), (0, 4), colspan=1, rowspan=1)  # Helix Angles + AngleCC
        ax8 = michelixtrace_plot.plt.subplot2grid((2, 5), (1, 4), colspan=1, rowspan=1)  # 2DCC
    
        cc_im = ax1.imshow(overlap_cc, cmap='jet', origin='lower', interpolation='nearest')
        ax1 = self.remove_ticks_and_scale_correctly(ax1, mic_np)
        ax1.set_title('Overlapping Cross Correlation %sx%s' % overlap_cc.shape, fontsize=4)
        cax = fig.add_axes([0.005, 0.65, 0.01, 0.20])
        cbar = fig.colorbar(cc_im, cax)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(4)
    
        ax2.set_title('Thresholded at alpha-value of {0}'.format(a_threshold), fontsize=4)
        bin_mask = np.ma.masked_where(binary < 0.5, -1 * binary)
        ax2.imshow(bin_mask, cmap='gray', origin='lower', interpolation='nearest')
        ax2 = self.remove_ticks_and_scale_correctly(ax2, mic_np)
        skel_mask = np.ma.masked_where(skel < 0.5, -1 * skel)
        ax2.imshow(skel_mask, cmap='autumn', origin='lower', interpolation='nearest')
    
        ax2_hist.set_title('Overlapping Cross Correlation Histogram', fontsize=4)
        ax2_hist.set_yscale('log', basey=10)
        n, bins, patches = ax2_hist.hist(overlap_cc[overlap_cc > cutoff].flatten(), 100, facecolor='green',
                                         linewidth=0.5, normed=True, log=True)
        ax2_hist.tick_params(axis='both', which='major', labelsize=4)
        x = np.linspace(overlap_cc.min(), overlap_cc.max(), 500)
        ax2_hist.plot(x, fitfunct.pdf(x, *fitparams), color='red')  # 1/float(bins[1]-bins[0]) *
        ax2_hist.set_ylim((10e-4, 1 / float(bins[1] - bins[0])))
    
        ax2_hist_trans.set_title('Overlapping Cross Correlation Histogram', fontsize=4)
        n, bins, patches = ax2_hist_trans.hist(overlap_cc[overlap_cc > cutoff].flatten(), 100, facecolor='green',
                                               linewidth=0.5, normed=True)
        ax2_hist_trans.tick_params(axis='both', which='major', labelsize=4)
        x = np.linspace(overlap_cc.min(), overlap_cc.max(), 500)
        ax2_hist_trans.plot(x, fitfunct.pdf(x, *fitparams), color='red')  # 1/float(bins[1]-bins[0]) *
        ax2_hist_trans.set_ylim((0, 1 / float(bins[1] - bins[0]) / 40.))
    
        ax3.imshow(mic_np, cmap='gray', origin='lower', interpolation='nearest')
        ax3.set_title('Micrograph with detected helices %sx%s' % mic_np.shape, fontsize=4)
        colors = plt.cm.rainbow(np.linspace(0, 1, len(helices)))[:, 0:3] / 2.0
        font = {'weight': 'bold', 'size': 3}
        for pers, c, (each_xcoord, each_ycoord) in zip(pers_lengths, colors, helices):
            ax3.plot(each_xcoord, each_ycoord, 'o', markersize=0.6, alpha=1, markeredgewidth=0.0, color=c)
            c_dark = c / 1.5
            xmean, ymean = each_xcoord.mean(), each_ycoord.mean()
            ax3.text(xmean, (ymean + each_ycoord.max()) / 2.0, "%.1f" % pers, color=c_dark, fontdict=font)
    
        ax3 = self.remove_ticks_and_scale_correctly(ax3, mic_np)
    
        ax4.imshow(ref, cmap='gray', origin='lower', interpolation='nearest')
        ax4.set_aspect('equal')
        ax4.set_title('Reference Helix', fontsize=4)
        ax4 = self.remove_ticks_and_scale_correctly(ax4, ref)
    
        cc_vals = ax5.imshow(cross_corr.T, origin='lower', interpolation='nearest')
        ax5.set_title('2D Cross-Corr %sx%s' % cross_corr.shape, fontsize=4)
        ax5.set_xticks([])
        ax5.set_yticks([])
        cax = fig.add_axes([0.77, 0.65, 0.01, 0.20])
        cbar = fig.colorbar(cc_vals, cax)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(4)
    
        reconst = ax6.imshow(overlap_cc, cmap='jet', origin='lower',
                             interpolation='nearest')  # extent=[-0.5, n-0.5, -0.5, m-0.5]
        X = np.array([i[0] for i in xy_center_grid.ravel()]).reshape(xy_center_grid.shape)
        Y = np.array([i[1] for i in xy_center_grid.ravel()]).reshape(xy_center_grid.shape)
        transparency = cross_corr - cross_corr.min()
        transparency = transparency / transparency.max()
        rgba_colors = [(1, 1, 1, a) for a in transparency.flatten()]
        ax6.scatter(X.flatten(), Y.flatten(), color=rgba_colors, s=0.5, linewidth=0)
        a1 = np.cos(np.deg2rad(angles)) * rho - X  # Vector AB from line-defining-point to all matrix points
        a2 = np.sin(np.deg2rad(angles)) * rho - Y  # Vector AB from line-defining-point to all matrix points
        AB = np.dstack([a1, a2])
        line_direction = np.dstack([np.cos(np.deg2rad(angles)), np.sin(np.deg2rad(angles))])
        length = np.einsum('ijk,ijk->ij', line_direction, AB)  # Dot product along last axis
        u = np.cos(np.deg2rad(angles)) * length / float(overlap_cc.shape[1])
        v = np.sin(np.deg2rad(angles)) * length / float(overlap_cc.shape[0])
        angles_plt = ax6.quiver(X, Y, u, v, linewidths=0.1, scale=1.0, alpha=0.5, color='white')
        ax6.set_title('Projection by Rho and Theta %sx%s' % angles.shape, fontsize=4)
        ax6.set_aspect(1.)
        ax6 = self.remove_ticks_and_scale_correctly(ax6, overlap_cc)
        ax6.set_xlim([-1, overlap_cc.shape[1] + 1])
        ax6.set_ylim([-1, overlap_cc.shape[0] + 1])
        cax = fig.add_axes([0.77, 0.25, 0.01, 0.20])
        cbar = fig.colorbar(reconst, cax)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(4)
    
        x, y = np.arange(peaks.shape[0]), np.arange(peaks.shape[1])
        X, Y = np.meshgrid(x, y)
        peaks_plt = ax7.imshow(peaks.T, origin='lower', interpolation='nearest')
        u, v = np.cos(np.deg2rad(angles.T - 90)), np.sin(np.deg2rad(angles.T - 90))
        angles_plt = ax7.quiver(X, Y, u, v, linewidths=0.2, pivot='mid')
        ax7.set_title('Helix Angles and AngleCC-Scores %sx%s' % peaks.shape, fontsize=4)
        ax7.set_xticks([])
        ax7.set_yticks([])
        ax7.set_xlim(-0.5, peaks.shape[0] - 0.5)
        ax7.set_ylim(-0.5, peaks.shape[1] - 0.5)
        cax = fig.add_axes([0.971, 0.65, 0.01, 0.20])
        cbar = fig.colorbar(peaks_plt, cax)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(4)
    
        rho_plot = ax8.imshow(cross_corr.T, origin='lower', interpolation='nearest')
        ax8.set_title('2D Cross-Corr %sx%s' % rho.shape, fontsize=4)
        ax8.set_xticks([])
        ax8.set_yticks([])
        cax = fig.add_axes([0.971, 0.25, 0.01, 0.20])
        cbar = fig.colorbar(rho_plot, cax)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(4)
    
        plt.subplots_adjust(left=0.04, right=0.96, bottom=0.18, top=0.92)
    
        fig.savefig(outfile[:-4] + '_verbose.pdf', dpi=600)
    
        return outfile
    
    

    def add_ccmap(self, overlap_cc, cross_corr, rho, angles, xy_center_grid, fig, ax1):
        ax1.set_title('Projection by Rho and Theta %sx%s' % angles.shape, fontsize=5)

        reconst = ax1.imshow(overlap_cc, cmap='jet', origin='lower', interpolation='nearest') # extent=[-0.5, n-0.5, -0.5, m-0.5]
        X = np.array([i[0] for i in xy_center_grid.ravel()]).reshape(xy_center_grid.shape)
        Y = np.array([i[1] for i in xy_center_grid.ravel()]).reshape(xy_center_grid.shape)
        transparency = cross_corr - cross_corr.min()
        transparency = transparency / transparency.max()
        rgba_colors = [(1, 1, 1, a) for a in transparency.flatten()]
        ax1.scatter(X.flatten(), Y.flatten(), color=rgba_colors, s=0.5, linewidth=0)
        a1 = np.cos(np.deg2rad(angles)) * rho - X # Vector AB from line-defining-point to all matrix points
        a2 = np.sin(np.deg2rad(angles)) * rho - Y # Vector AB from line-defining-point to all matrix points
        AB = np.dstack([a1, a2])
        line_direction = np.dstack([np.cos(np.deg2rad(angles)), np.sin(np.deg2rad(angles))])
        length = np.einsum('ijk,ijk->ij', line_direction, AB) # Dot product along last axis
        u = np.cos(np.deg2rad(angles)) * length / float(overlap_cc.shape[1])
        v = np.sin(np.deg2rad(angles)) * length / float(overlap_cc.shape[0])
        angles_plt = ax1.quiver(X, Y, u, v, linewidths=0.1, scale=1.0, alpha=0.5, color='white')
        ax1.set_aspect(1.)
        ax1 = self.remove_ticks_and_scale_correctly(ax1, overlap_cc)
        ax1.set_xlim([-1, overlap_cc.shape[1] + 1])
        ax1.set_ylim([-1, overlap_cc.shape[0] + 1])
        cax = fig.add_axes([0.003, 0.695, 0.01, 0.22])
        cbar = fig.colorbar(reconst, cax)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(4)
        
        return ax1


    def add_referene_image(self, ref, ax2):
        ax2.set_title('Reference Helix', fontsize=5)
        
        ax2.imshow(ref, cmap='gray', origin='lower', interpolation='nearest')
        ax2.set_aspect('equal')
        ax2 = self.remove_ticks_and_scale_correctly(ax2, ref)

        return ax2


    def add_thresholded_ccmap(self, binary, skel, a_threshold, absolute_threshold, mic_np, ax3):
        if absolute_threshold is None:
            title = 'Thresholded at alpha-value of {0}'.format(a_threshold)
        else:
            title = 'Thresholded at cc-threshold of {0}'.format(absolute_threshold)

        ax3.set_title(title, fontsize=5)

        bin_mask = np.ma.masked_where(binary < 0.5, -1 * binary)
        ax3.imshow(bin_mask, cmap='gray', origin='lower', interpolation='nearest')
        ax3.plot(None, label='thresholded', color='black')
        ax3 = self.remove_ticks_and_scale_correctly(ax3, mic_np)

        skel_mask = np.ma.masked_where(skel < 0.5, -1 * skel)
        ax3.imshow(skel_mask, cmap='autumn', origin='lower', interpolation='nearest')
        ax3.plot(None, label='cleaned up', color='red')
        ax3.legend(fontsize=4)

        return ax3


    def add_cc_histogram(self, overlap_cc, lamb, absolute_threshold, background_cutoff,  a_threshold, ax4):
        ax4.set_title('Projection Map Histogram & Thresholding', fontsize=5)
        
        n, bins, patches = ax4.hist(overlap_cc[overlap_cc>background_cutoff].flatten(), 80, facecolor='grey',
                                    linewidth=0.0, normed=True)
        ax4.tick_params(axis='both', which='major', labelsize=4)
        x = np.linspace(overlap_cc.min(), overlap_cc.max(), 500)

        if absolute_threshold is None:
            fitfunct = stats.expon
            fitparams = [0, lamb]

            ax4.plot(x, fitfunct.pdf(x, *fitparams), color='red', linewidth=1,
                     label='Exp. Fit\nlambda=%.3f'%lamb)  # 1/float(bins[1]-bins[0]) *

            absolute_threshold = fitfunct.ppf(1 - a_threshold, *fitparams)
            ax4.axvline(absolute_threshold, color='green', linewidth=1, label="alpha=%.2e" % a_threshold)

            colors_choice = {}
            colors_choice[-1.] = 'blue'
            colors_choice[1.] = 'orange'
            for each_d in colors_choice:
                alt_alpha = a_threshold * (10.0 ** each_d)
                alt_absolute_threshold = fitfunct.ppf(1 - (alt_alpha), *fitparams)
                ax4.axvline(alt_absolute_threshold, color=colors_choice[each_d], linewidth=0.5, linestyle=':', label="alpha=%.2e" % alt_alpha)

            ax4.text(1.1 * absolute_threshold, 1 / float(bins[1] - bins[0]) / 20 * 0.9,
                     "cc_threshold=%.3e\nalpha=%.2e" % (absolute_threshold, a_threshold),
                     color='green', fontsize=5)

        else:
            ax4.axvline(absolute_threshold, color='green', linewidth=1)
            ax4.text(1.1 * absolute_threshold, 1 / float(bins[1] - bins[0]) / 20 * 0.9,
                     "cc threshold=%.3e" % (absolute_threshold),
                     color='green', fontsize=5)

        ax4.set_ylim((0, 1 / float(bins[1] - bins[0]) / 20.))
        for b, p in zip(bins, patches):
            if b > absolute_threshold:
                plt.setp(p, 'facecolor', 'green')
        
        legend = ax4.legend(fontsize=5, labelspacing=0.2, loc=7)
        
        return ax4


    def add_micrograph(self, ax5, mic_np, helices, pers_lengths):
        ax5.set_title('Micrograph with detected helices %sx%s. Numbers denote persistence length in um' % mic_np.shape,
                      fontsize=5)

        ax5.imshow(mic_np, cmap='gray', origin='lower', interpolation='nearest')
        colors = plt.cm.hsv(np.linspace(0, 1, len(helices)))[:, 0:3]  # / 1.0
        font = {'weight': 'bold', 'size': 5}
        for pers, c, (each_xcoord, each_ycoord) in zip(pers_lengths, colors, helices):
            ax5.plot(each_xcoord, each_ycoord, 'o', markersize=2, alpha=1, markeredgewidth=0.0, color=c)
            c_dark = c / 5.0
            xmean, ymean = each_xcoord.mean(), each_ycoord.mean()
            ax5.text(xmean, ymean, "%.1f" % pers, color=c_dark, fontdict=font)
        ax5 = self.remove_ticks_and_scale_correctly(ax5, mic_np)
        
        return ax5 

    
    def visualize_traces_in_diagnostic_plot(self, infile, outfile, overlap_cc, binary, helices, mic, ref, cross_corr,
                                            rho, angles, xy_center_grid, lamb, absolute_threshold, background_cutoff,
                                            skel, pers_lengths, feature_set, a_threshold):
        mic.read_image(infile)
        mic_np = np.copy(EMNumPy.em2numpy(mic))
        michelixtrace_plot = DiagnosticPlot()
        fig = michelixtrace_plot.add_header_and_footer(feature_set, infile, outfile)
    
        gs = gridspec.GridSpec(2, 4, width_ratios=[1, 1, 1.25, 1.25], height_ratios=[1, 1.7])
    
        ax1 = fig.add_subplot(gs[0, 0])  # Reconstruction
        ax2 = fig.add_subplot(gs[1, 1])  # Reference
        ax3 = fig.add_subplot(gs[0, 1])  # Thresholded
        ax4 = fig.add_subplot(gs[1, 0])  # Histogram
        ax5 = fig.add_subplot(gs[:, 2:])  # Micrograph
    
        ax1 = self.add_ccmap(overlap_cc, cross_corr, rho, angles, xy_center_grid, fig, ax1)
        ax2 = self.add_referene_image(ref, ax2)
        ax3 = self.add_thresholded_ccmap(binary, skel, a_threshold, absolute_threshold, mic_np, ax3)
        ax4 = self.add_cc_histogram(overlap_cc, lamb, absolute_threshold, background_cutoff, a_threshold, ax4)
        ax5 = self.add_micrograph(ax5, mic_np, helices, pers_lengths)
    
        fig.tight_layout()
        fig.subplots_adjust(left=0.045, right=0.99, bottom=0.21, top=0.93)
        fig.savefig(outfile, dpi=600)
    
        return outfile



#!/usr/bin/env python
"""
Test module to check segmentclass
"""
from EMAN2 import EMData, EMUtil
from fundamentals import rot_shift2D
from glob import glob
from multiprocessing import cpu_count
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.segment2d.segmentclass import SegmentClass, SegmentClassPar
from spring.tests.csinfrastr.test_csreadinput import EndToEndTest
from spring.tests.segment2d.test_segmentexam import HelixProjectionPreparation
import numpy as np
import os
import random
import shutil


class HelixClassPreparation(HelixProjectionPreparation):
    def shift_segments(self):
        img = EMData()
        
        for each_img_id in list(range(EMUtil.get_image_count('test_segments.hdf'))):
            img.read_image('test_segments.hdf', each_img_id)
            img = rot_shift2D(img, random.randrange(-5, 5), random.randrange(-3, 3), random.randrange(-3, 3))
            img.write_image('test_segments.hdf', each_img_id)
            
            
    def prepare_helix_classes(self, hlx_cls_count=3):
        """
        * Function to generate a projection of 4 classes
        """
        self.number_of_classes = hlx_cls_count
        for class_helix in range(self.number_of_classes):
            self.width_of_helix_in_angstrom = 250 + 100 * class_helix
            self.prepare_helix_projection()
            self.shift_segments()
        

    def prepare_custom_filter_file(self, custom_filter_file, dimension, binfactor=1):
        number_of_frequencies = (dimension/2)/binfactor
#        number_of_frequencies = (self.helix_micrograph.get_xsize()/2)/binfactor
        frequencies = np.linspace(0, 0.5, number_of_frequencies)
        ffile = open(custom_filter_file, 'w')
        for each_frequency in frequencies:
            filter_coefficient = np.sin(np.degrees(each_frequency))
            if filter_coefficient > 1:
                filter_coefficient = 1
            elif filter_coefficient < 0:
                filter_coefficient = 0
            
            ffile.write('{0}\t{1}\n'.format(each_frequency, filter_coefficient))
        
        ffile.close()


    def setup_helix_or_particle_dimensions(self):
        self.feature_set.parameters['Estimated helix width and height in Angstrom'] = ((self.width_of_helix_in_angstrom,
        int(1.5 * self.width_of_helix_in_angstrom)))

    def setup_common_non_mpi_parameters(self):
        self.prepare_helix_classes()
        self.custom_filter_file = 'test_filter_file.dat'
        self.test_bin_factor = 2
        self.prepare_custom_filter_file(self.custom_filter_file, self.test_bin_factor)
        self.logfile, self.directory = EndToEndTest().define_logfile_and_directory(self.feature_set.progname)
        self.feature_set.logfile = self.logfile
        self.feature_set.parameters['Image input stack'] = 'test_segments.hdf'
        self.feature_set.parameters['Class average stack'] = 'test_averages.hdf'
        
        self.feature_set.parameters['Spring database option']=False
        self.feature_set.parameters['Database prepare option']=False
        self.feature_set.parameters['spring.db file']=os.path.join(os.pardir, 'spring.db')
            
        self.feature_set.parameters['Reference image option']=False
        self.feature_set.parameters['Image reference stack']='test_ref_segments.hdf'
        self.feature_set.parameters['Class variance stack'] = 'test_variances.hdf'
        self.feature_set.parameters['Eigenimage stack'] = 'test_eigenimages.hdf'
        self.feature_set.parameters['Pixel size in Angstrom'] = float(self.pixelsize)
        self.setup_helix_or_particle_dimensions()
        self.feature_set.parameters['Binning option'] = True
        self.feature_set.parameters['Binning factor'] = self.test_bin_factor
        self.feature_set.parameters['Number of classes'] = int(self.number_of_classes)
        self.feature_set.parameters['Number of iterations'] = int(3)
        self.feature_set.parameters['Keep intermediate files']=False
        
        self.feature_set.parameters['Limit in-plane rotation'] = True
        self.feature_set.parameters['Delta in-plane rotation angle'] = 10.0
        self.feature_set.parameters['X and Y translation range in Angstrom'] = (20, 15)
        self.feature_set.parameters['High-pass filter option'] = False
        self.feature_set.parameters['Low-pass filter option'] = False
        self.feature_set.parameters['High and low-pass filter cutoffs in 1/Angstrom'] = (0.01, 0.06)
        self.feature_set.parameters['Custom filter option'] = False
        self.feature_set.parameters['Custom-built filter file'] = self.custom_filter_file
        self.feature_set.parameters['Automatic filter option'] = False
        self.feature_set.parameters['B-Factor']=0
        self.feature_set.parameters['Temporary directory']=os.path.abspath(os.curdir)


class TestSegmentClassClear(object):
    def clear_segment_files(self):
        segment_files = glob('test_segments*.hdf')
        for segment_file in segment_files:
            os.remove(segment_file)
        pass

    def clear_logfiles(self):
        try:
            os.remove(os.path.join(os.pardir, self.logfile))
        except:
            pass
        
        logfiles = glob('logfile*')
        for logfile in logfiles:
            os.remove(logfile)

    def clear_directories_including_content(self):
        directories = ['sxk_means*', 'segmentalign*', 'particlealign*']
        
        for directory in directories:
            cluster_directory_list = glob(directory)
            for cluster_directory in cluster_directory_list:
                all_files_in_directory = os.listdir(cluster_directory)
                for each_file in all_files_in_directory:
                    os.remove(os.path.join(cluster_directory, each_file))
                os.rmdir(cluster_directory)


    def teardown_segmentclass(self):
        self.clear_segment_files()
        os.remove(self.custom_filter_file)
        os.remove('spring.db')
        os.remove('test_averages.hdf')
        os.remove('test_variances.hdf')
        os.remove('test_eigenimages.hdf')
        self.clear_logfiles()
        self.clear_directories_including_content()


class TestSegmentClass(HelixClassPreparation, SegmentClass, TestSegmentClassClear):

    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()

        self.feature_set = SegmentClassPar()
        self.setup_common_non_mpi_parameters()
        
        self.feature_set.parameters['MPI option']=False
        self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())

        super(TestSegmentClass, self).__init__(self.feature_set)


    def teardown(self):
        self.teardown_segmentclass()
        
        self.testingdir.remove()


class TestSegmentClassMain(TestSegmentClass):
    def do_test_case_sc1(self):
        """
        * Standard single-CPU segmentclass test
        """
        self.classify()


class TestSegmentClassMore(TestSegmentClass):
    def do_test_case_sc2_reference_image(self):
        """
        * Segmentclass: Test reference image option
        """
        self.feature_set.parameters['Reference image option']=True
 
        shutil.copy('test_segments.hdf', 'test_ref_segments.hdf')
        super(TestSegmentClass, self).__init__(self.feature_set)
        self.classify()
        os.remove('test_ref_segments.hdf')
        os.remove('test_ref_segments2binned.hdf')
 
     
    def do_test_case_sc3_single_iteration(self):
        """
        * Segmentclass single iteration
        """
        self.feature_set.parameters['Number of iterations'] = 0
 
        super(TestSegmentClass, self).__init__(self.feature_set)
        self.classify()
         

    def do_test_case_sc7_many_classes(self):
        """
        * Segmentclass larger number of classes; splitting of sxk_means images
        """
        self.test_cls_count = 7

        os.remove('test_segments.hdf')
        self.prepare_helix_classes(self.test_cls_count)
        self.feature_set.parameters['Number of classes'] = self.test_cls_count
        super(TestSegmentClass, self).__init__(self.feature_set)
        self.classify()


class TestSegmentClassEndToEnd(TestSegmentClass):
    def do_end_to_end_test_sc_inputfile(self):
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)

    def do_end_to_end_test_sc_prompt(self):
        EndToEndTest().do_end_to_end_prompt(self.feature_set)

class TestSegmentClassMpi(TestSegmentClass):
    def do_end_to_end_test_sc_inputfile_MPI(self):

        self.feature_set.parameters['MPI option']=True
#        self.feature_set.parameters['Number of CPUs']=2
        self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())
        
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)


def main():
    tsce = TestSegmentClassMain()
    tsce.setup()
    tsce.do_test_case_sc1()
        
if __name__ == '__main__':
    main()



# Author: Carsten Sachse 21-Sep-2010
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
""" 
Program to evaluate the performance of scanner by correlating adjacent rows 
and lines with each other from a pure noise image.
"""
from EMAN2 import Util
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import DiagnosticPlot
from spring.csinfrastr.csreadinput import OptHandler
from spring.micprgs.scansplit import Micrograph
from statistics import ccc

class ScanRowColCorrPar:
    """
    Class to initiate default dictionary with input parameters including help and range values and 
    status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'scanrowcolcorr'
        self.proginfo = __doc__
        self.code_files = [self.progname]

        self.scanrowcolcorr_features = Features()
        self.feature_set = self.scanrowcolcorr_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    
    def define_parameters_and_their_properties(self):
        self.feature_set = self.scanrowcolcorr_features.set_inp_micrograph(self.feature_set)
        self.feature_set = self.scanrowcolcorr_features.set_output_plot(self.feature_set, self.progname + '_diag.pdf')
    
        self.feature_set = self.set_mic_area(self.feature_set)
    
    def define_program_states(self):
        self.feature_set.program_states['readmic']='Read input micrograph'
        self.feature_set.program_states['evalrowcc']='Evaluate row-to-row cross-correlation'
        self.feature_set.program_states['evalcolcc']='Evaluate column-to-column cross-correlation'
        self.feature_set.program_states['visrowcolcc']='Visualize row-to-row and column-to-column correlation'

    def set_mic_area(self, feature_set):
        inp3 = 'Percentage of micrograph area to be analyzed'
        feature_set.parameters[inp3] = int(90)
        feature_set.hints[inp3] = '0-100 percent from the row and column dimensions of micrograph.'
        feature_set.properties[inp3] = feature_set.Range(0, 100, 1)
        feature_set.level[inp3]='expert'
        
        return feature_set


class ScanRowColCorr(object):
    """
    * Class that holds all functions required for computing row-to-row and column-to-column cross-correlation

    * __init__ Function to read in the entered parameter dictionary and load micrograph

    #. Usage: ScanRowColCorr(pardict)
    #. Input: pardict = OrderedDict of program parameters

    """
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            p = self.feature_set.parameters
            self.infile=p['Micrograph']
            self.outfile=p['Diagnostic plot']
            self.area=p['Percentage of micrograph area to be analyzed']

            #Micrograph.readmic(self, self.infile)
            self.img, self.nx, self.ny = Micrograph().readmic(self.infile)

    def evalrowcc(self, area = None):
        """
        * Function to evaluate cross-correlation of adjacent rows

        #. Usage: rows, ccrows, output2 = evalrowcc(area)
        #. Input: area = area in percent to included in analysis of input micrograph, i.e. exclude label
        #. Output: list of rows, list of cross-correlation

        """
        if area is None: area = self.area
        self.log.fcttolog()
        edge = int(self.nx*(100 - area)/100)

        self.row = []
        self.ccrow = []
        k = 0 
        for eachRow in range(1 + edge, self.ny - 1 - edge):
            if (eachRow == 1 + edge): 
                img1dline1 = self.img.get_row(eachRow)
            else:
                img1dline2 = self.img.get_row(eachRow)
                #img1dline2.write_image('rowstackcc.spi', k)
                k = k +1 
                self.row.append(eachRow)
                self.ccrow.append(ccc(img1dline1, img1dline2))
                img1dline1 = img1dline2
                self.log.ilog('Correlated row %d with row %d: cross-correlation of %f' %
                              (eachRow, eachRow -1, self.ccrow[-1]))
        return self.row, self.ccrow

    def evalcolcc(self, area = None):
        """
        * Function to evaluate cross-correlation of adjacent columns

        #. Usage: cols, cccols, output2 = evalrowcc(area)
        #. Input: area = area in percent to included in analysis of input micrograph, i.e. exclude label
        #. Output: list of columns, list of cross-correlation
        """
        if area is None: area = self.area
        self.log.fcttolog()

        self.col = []
        self.cccol = []
        for eachCol in range(0, self.nx):
            if (eachCol == 0): 
                img1dline1 = self.img.get_col(eachCol)
                img1dline1 = Util.window(img1dline1, int(area*self.ny/100), 1, 1, 0, 0, 0)
            else:
                img1dline2 = self.img.get_col(eachCol)
                img1dline2 = Util.window(img1dline2, int(area*self.ny/100), 1, 1, 0, 0, 0)
                self.col.append(eachCol)
                self.cccol.append(ccc(img1dline1, img1dline2))
                img1dline1 = img1dline2
                self.log.ilog('Correlated column %d with column %d: cross-correlation of %f' %
                              (eachCol, eachCol -1, self.cccol[-1]))
        return self.col, self.cccol

    def visrowcolcc(self, row = None, ccrow = None, col = None, cccol = None):
        """
        * Function to visualize results of row-to-row and column-to-column cross-correlation \
        using matplotlib

        #. Usage: output = visrowcolcc(row, ccrow, col, cccol)
        #. Input: row = list of rows, ccrow = list of row-to-row cross-correlation, col = list of \
                        columns, cccol = list of column-to-column correlation
        #. Output: output plot to saved to PDF, SVG, or PNG, TIF, JPG format
        """
        self.log.fcttolog()

        if row is None: row = self.row
        if ccrow is None: ccrow = self.ccrow
        if col is None: col = self.col
        if cccol is None: cccol = self.cccol

        scanrowcolcorr_fit_plot = DiagnosticPlot()
        self.fig = scanrowcolcorr_fit_plot.add_header_and_footer(self.feature_set)
        
        ax1 = self.fig.add_subplot(111)
        ax1.plot(row, ccrow, 'x', markeredgewidth=0.4, markersize=0.4, label = 'row-to-row correlation')
        ax1.plot(col, cccol, 'x', markeredgewidth=0.4, markersize=0.4, label = 'column-to-column correlation')
        ax1.set_title('Adjacent pixel correlation of pure noise image')
        ax1.legend(loc=3)
        ax1.set_xlabel('Row or column number')
        ax1.set_ylabel('Cross-correlation')
        ax1.set_ylim(0, 1)
        ax1.set_xlim(0, row[-1])
        ax1.minorticks_on()
        ax1.grid(True)
        self.fig.savefig(self.outfile)

        return self.fig


    def perform_scanrowcolcorr(self):
        self.log.plog(10)
        self.evalrowcc()
        self.log.plog(40)
        self.evalcolcc()
        self.log.plog(70)
        self.visrowcolcc()
        self.log.endlog(self.feature_set)
        

def main():
    # Option handling
    parset = ScanRowColCorrPar()
    mergeparset = OptHandler(parset)

    ###### Program
    mic = ScanRowColCorr(mergeparset)
    mic.perform_scanrowcolcorr()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Test module to check segclasslayer
"""
import os
from statistics import add_series
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.segment2d.segclasslayer import SegClassLayer, SegClassLayerPar
from spring.tests.segment2d.test_segmentexam import HelixProjectionPreparation
from spring.tests.csinfrastr.test_csreadinput import EndToEndTest


class HelixSingleClassPreparation(HelixProjectionPreparation):
    def prepare_helix_class(self):
        """
        * Function to generate a projection of 4 classes
        """
        self.width_of_helix_in_angstrom = 200
        self.prepare_helix_projection()
        self.helix_class_average, self.helix_class_variance = add_series('test_segments.hdf')
        self.class_average = 'test_class_avg_segments.hdf'
        self.helix_class_average.append_image(self.class_average)
        self.class_variance = 'test_class_var_segments.hdf'

class TestSegClassLayer(HelixSingleClassPreparation, SegClassLayer):
    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()

        self.feature_set = SegClassLayerPar()
        self.prepare_helix_class()

        self.feature_set.parameters['Class average stack']=self.class_average
        self.feature_set.parameters['Batch mode']=True#
        self.feature_set.parameters['Diagnostic plot']='test_segclasslayer.png'
        self.feature_set.parameters['Class format']='real'
        self.feature_set.parameters['Pixel size in Angstrom']=float(self.pixelsize)
        self.feature_set.parameters['Precise helix width in Angstrom']=int(1.1*self.width_of_helix_in_angstrom)
        self.feature_set.parameters['B-Factor']=-10000
        self.feature_set.parameters['Power spectrum resolution cutoff in 1/Angstrom']=float(1/(self.pixelsize*8))
        self.feature_set.parameters['Class number range to be analyzed']=((0, 0))
        self.feature_set.parameters['Layer line positions']='0.014, 0.029'
        self.feature_set.parameters['Pad option']=bool(True)

        super(TestSegClassLayer, self).__init__(self.feature_set)


    def teardown(self):
        os.remove('test_segments.hdf')
        os.remove(self.class_average)

        if os.path.splitext(self.feature_set.parameters['Diagnostic plot'])[-1].endswith('pdf'):
            os.remove('test_segclasslayer.pdf')
        else:
            os.remove('test_segclasslayer.png')
        os.remove('test_class_avg_segments4binned.hdf')

        self.testingdir.remove()


class TestSegClassLayerMain(TestSegClassLayer):
    def do_test_case_scl1(self):
        self.extract_and_visualize_layer_lines()


class TestSegClassLayerMore(TestSegClassLayer):
    def do_test_case_scl2(self):
        self.prepare_helix_class()
        self.feature_set.parameters['Class number range to be analyzed']=((0, 100))
        self.feature_set.parameters['Diagnostic plot']='test_segclasslayer.pdf'
        super(TestSegClassLayer, self).__init__(self.feature_set)
        
        self.extract_and_visualize_layer_lines()
        

class TestSegClassLayerEndToEnd(TestSegClassLayer):
    def do_end_to_end_test_scl_inputfile(self):
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)

    def do_end_to_end_test_scl_prompt(self):
        EndToEndTest().do_end_to_end_prompt(self.feature_set)

def main():
    tscl = TestSegClassLayerMain()
    tscl.setup()
    tscl.do_test_case_scl1()
        
if __name__ == '__main__':
    main()
